#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
addrlist: Address list classes for the MMGen suite
"""

from .util import suf, make_chksum_N, Msg, die
from .objmethods import MMGenObject, HiliteStr, InitErrors
from .obj import MMGenListItem, ListItemAttr, MMGenDict, TwComment, WalletPassword
from .key import PrivKey
from .addr import MMGenID, MMGenAddrType, CoinAddr, AddrIdx, AddrListID, ViewKey

class AddrIdxList(tuple, InitErrors, MMGenObject):

	max_len = 1000000

	def __new__(cls, *, fmt_str=None, idx_list=None, sep=','):
		try:
			if fmt_str:
				def gen():
					for i in (fmt_str.split(sep)):
						j = [int(x) for x in i.split('-')]
						if len(j) == 1:
							yield j[0]
						elif len(j) == 2:
							if j[0] > j[1]:
								raise ValueError(f'{i}: invalid range')
							yield from range(j[0], j[1] + 1)
						else:
							raise ValueError(f'{i}: invalid range')
				idx_list = tuple(gen())
			return tuple.__new__(cls, sorted({AddrIdx(i) for i in (idx_list or [])}))
		except Exception as e:
			return cls.init_fail(e, idx_list or fmt_str)

	@property
	def id_str(self):

		def gen():
			i_save = self[0]
			yield f'{i_save}'
			in_range = False

			for i in self[1:]:
				if i - i_save == 1:
					in_range = True
				else:
					if in_range:
						in_range = False
						yield f'-{i_save}'
					yield f',{i}'
				i_save = i

			if in_range:
				yield f'-{i_save}'

		return ''.join(gen()) if self else ''

class AddrListEntryBase(MMGenListItem):
	invalid_attrs = {'proto'}
	def __init__(self, proto, **kwargs):
		self.__dict__['proto'] = proto
		MMGenListItem.__init__(self, **kwargs)

class AddrListEntry(AddrListEntryBase):
	addr          = ListItemAttr(CoinAddr, include_proto=True)
	addr_p2pkh    = ListItemAttr(CoinAddr, include_proto=True)
	idx           = ListItemAttr(AddrIdx) # not present in flat addrlists
	comment       = ListItemAttr(TwComment, reassign_ok=True)
	sec           = ListItemAttr(PrivKey, include_proto=True)
	viewkey       = ListItemAttr(ViewKey, include_proto=True)
	wallet_passwd = ListItemAttr(WalletPassword)

class AddrListChksum(HiliteStr):
	color = 'pink'
	trunc_ok = False

	def __new__(cls, addrlist):
		ea = addrlist.al_id.mmtype.extra_attrs or () # add viewkey and passwd to the mix, if present
		lines = [' '.join(
					addrlist.chksum_rec_f(e) +
					tuple(getattr(e, a) for a in ea if getattr(e, a))
				) for e in addrlist.data]
		return str.__new__(cls, make_chksum_N(' '.join(lines), nchars=16, sep=True))

class AddrListIDStr(HiliteStr):
	color = 'green'
	trunc_ok = False

	def __new__(cls, addrlist, *, fmt_str=None):
		idxs = [e.idx for e in addrlist.data]
		prev = idxs[0]
		ret = [prev]
		for i in idxs[1:]:
			if i == prev + 1:
				if i == idxs[-1]:
					ret.extend(['-', i])
			else:
				if prev != ret[-1]:
					ret.extend(['-', prev])
				ret.extend([',', i])
			prev = i
		s = ''.join(map(str, ret))

		if fmt_str:
			ret = fmt_str.format(s)
		else:
			proto = addrlist.proto
			coin = 'BTC' if proto.coin == 'BCH' and not addrlist.cfg.cashaddr else proto.coin
			mmtype = addrlist.al_id.mmtype
			ret = '{}{}{}[{}]'.format(
				addrlist.al_id.sid,
				(f'-{coin}', '')[coin == 'BTC'],
				(f'-{mmtype}', '')[mmtype in ('L', 'E')],
				s)

		addrlist.dmsg_sc('id_str', ret[8:].split('[')[0])

		return str.__new__(cls, ret)

class AddrListData(list, MMGenObject):
	pass

class AddrList(MMGenObject): # Address info for a single seed ID
	entry_type   = AddrListEntry
	main_attr    = 'addr'
	desc         = 'address'
	gen_desc     = 'address'
	gen_desc_pl  = 'es'
	gen_addrs    = True
	gen_passwds  = False
	gen_keys     = False
	has_keys     = False
	chksum_rec_f = lambda foo, e: (str(e.idx), e.addr.views[e.addr.view_pref])

	def dmsg_sc(self, desc, data):
		Msg(f'sc_debug_{desc}: {data}')

	def noop(self, desc, data):
		pass

	def __init__(
			self,
			cfg,
			proto,
			*,
			infile    = '',
			al_id     = '',
			adata     = [],
			seed      = '',
			addr_idxs = '',
			src       = '',
			addrlist  = '',
			keylist   = '',
			mmtype    = None,
			key_address_validity_check = None, # None=prompt user, True=check without prompt, False=skip check
			skip_chksum = False,
			skip_chksum_msg = False,
			add_p2pkh = False):

		self.cfg = cfg
		self.ka_validity_chk = key_address_validity_check
		self.add_p2pkh = add_p2pkh
		self.proto = proto
		do_chksum = False

		if not cfg.debug_addrlist:
			self.dmsg_sc = self.noop

		if seed and addr_idxs:   # data from seed + idxs
			self.al_id = AddrListID(sid=seed.sid, mmtype=MMGenAddrType(proto, mmtype or proto.dfl_mmtype))
			src = 'gen'
			adata = self.generate(
				seed,
				addr_idxs if isinstance(addr_idxs, AddrIdxList) else AddrIdxList(fmt_str=addr_idxs))
			do_chksum = True
		elif infile:             # data from MMGen address file
			self.infile = infile
			adata = self.file.parse_file(infile) # sets self.al_id
			do_chksum = True
		elif al_id and adata:    # data from tracking wallet
			self.al_id = al_id
		elif addrlist:           # data from flat address list
			self.al_id = None
			from .util import remove_dups
			addrlist = remove_dups(addrlist, edesc='address', desc='address list')
			adata = AddrListData([AddrListEntry(proto=proto, addr=a) for a in addrlist])
		elif keylist:            # data from flat key list
			self.al_id = None
			from .util import remove_dups
			keylist = remove_dups(keylist, edesc='key', desc='key list', hide=True)
			adata = AddrListData([AddrListEntry(proto=proto, sec=PrivKey(proto=proto, wif=k)) for k in keylist])
		elif seed or addr_idxs:
			die(3, 'Must specify both seed and addr indexes')
		elif al_id or adata:
			die(3, 'Must specify both al_id and adata')
		else:
			die(3, f'Incorrect arguments for {type(self).__name__}')

		# al_id, adata now set
		self.data = adata
		self.num_addrs = len(adata)
		self.fmt_data = ''
		self.chksum = None

		if self.al_id is None:
			return

		if type(self) is ViewKeyAddrList:
			if not 'viewkey' in self.al_id.mmtype.extra_attrs:
				die(1, f'viewkeys not supported for address type {self.al_id.mmtype.desc!r}')

		self.id_str = AddrListIDStr(self)

		if type(self) is KeyList:
			return

		if do_chksum and not skip_chksum:
			self.chksum = AddrListChksum(self)
			if not skip_chksum_msg:
				self.do_chksum_msg(record=src=='gen')

	def do_chksum_msg(self, record):
		chk = 'Check this value against your records'
		rec = f'Record this checksum: it will be used to verify the {self.desc} file in the future'
		self.cfg._util.qmsg(
			f'Checksum for {self.desc} data {self.id_str.hl()}: {self.chksum.hl()}\n' +
			(chk, rec)[record])

	def generate(self, seed, addr_idxs):

		seed = self.scramble_seed(seed.data)
		self.dmsg_sc('seed', seed[:8].hex())

		mmtype = self.al_id.mmtype

		gen_wallet_passwd = type(self) in (KeyAddrList, ViewKeyAddrList) and 'wallet_passwd' in mmtype.extra_attrs
		gen_viewkey       = type(self) in (KeyAddrList, ViewKeyAddrList) and 'viewkey' in mmtype.extra_attrs

		if self.gen_addrs:
			from .keygen import KeyGenerator
			from .addrgen import AddrGenerator
			kg = KeyGenerator(self.cfg, self.proto, mmtype.pubkey_type)
			ag = AddrGenerator(self.cfg, self.proto, mmtype)
			if self.add_p2pkh:
				ag2 = AddrGenerator(self.cfg, self.proto, 'compressed')

		from .derive import derive_coin_privkey_bytes

		t_addrs = len(addr_idxs)
		le = self.entry_type
		out = AddrListData()
		CR = '\n' if self.cfg.debug_addrlist else '\r'

		for pk_bytes in derive_coin_privkey_bytes(seed, addr_idxs):

			if not self.cfg.debug:
				self.cfg._util.qmsg_r(
					f'{CR}Generating {self.gen_desc} #{pk_bytes.idx} ({pk_bytes.pos} of {t_addrs})')

			e = le(proto=self.proto, idx=pk_bytes.idx)

			e.sec = PrivKey(
				self.proto,
				pk_bytes.data,
				compressed  = mmtype.compressed,
				pubkey_type = mmtype.pubkey_type)

			if self.gen_addrs:
				data = kg.gen_data(e.sec)
				e.addr = ag.to_addr(data)
				if self.add_p2pkh:
					e.addr_p2pkh = ag2.to_addr(data)
				if gen_viewkey:
					e.viewkey = ag.to_viewkey(data)
				if gen_wallet_passwd:
					e.wallet_passwd = self.gen_wallet_passwd(
						e.viewkey.encode() if type(self) is ViewKeyAddrList else e.sec)
			elif self.gen_passwds:
				e.passwd = self.gen_passwd(e.sec) # TODO - own type

			out.append(e)

		self.cfg._util.qmsg('{}{}: {} {}{} generated{}'.format(
			CR,
			self.al_id.hl(),
			t_addrs,
			self.gen_desc,
			suf(t_addrs, self.gen_desc_pl),
			' ' * 15))

		return out

	def gen_wallet_passwd(self, privbytes):
		from .proto.btc.common import hash256
		return WalletPassword(hash256(privbytes)[:16].hex())

	def check_format(self, addr):
		return True # format is checked when added to list entry object

	def scramble_seed(self, seed):
		is_btcfork = self.proto.base_coin == 'BTC'
		if is_btcfork and self.al_id.mmtype == 'L' and not self.proto.testnet:
			self.dmsg_sc('str', '(none)')
			return seed
		if self.proto.base_coin == 'ETH':
			scramble_key = self.proto.coin.lower()
		else:
			scramble_key = (self.proto.coin.lower()+':', '')[is_btcfork] + self.al_id.mmtype.name
		from .crypto import Crypto
		if self.proto.testnet:
			scramble_key += ':' + self.proto.network
		self.dmsg_sc('str', scramble_key)
		return Crypto(self.cfg).scramble_seed(seed, scramble_key.encode())

	def idxs(self):
		return [e.idx for e in self.data]

	def addrs(self):
		return [f'{self.al_id.sid}:{e.idx}' for e in self.data]

	def addrpairs(self):
		return [(e.idx, e.addr) for e in self.data]

	def coinaddrs(self):
		return [e.addr for e in self.data]

	def comments(self):
		return [e.comment for e in self.data]

	def entry(self, idx):
		for e in self.data:
			if idx == e.idx:
				return e

	def coinaddr(self, idx):
		for e in self.data:
			if idx == e.idx:
				return e.addr

	def comment(self, idx):
		for e in self.data:
			if idx == e.idx:
				return e.comment

	def set_comment(self, idx, comment):
		for e in self.data:
			if idx == e.idx:
				e.comment = comment

	def make_reverse_dict_addrlist(self, coinaddrs):
		d = MMGenDict()
		b = coinaddrs
		for e in self.data:
			try:
				d[b[b.index(e.addr)]] = (MMGenID(self.proto, f'{self.al_id}:{e.idx}'), e.comment)
			except ValueError:
				pass
		return d

	def add_wifs(self, key_list):
		"""
		Match WIF keys in a flat list to addresses in self by generating all
		possible addresses for each key.
		"""
		def gen_addr(pk, t):
			at = self.proto.addr_type(t)
			from .keygen import KeyGenerator
			from .addrgen import AddrGenerator
			kg = KeyGenerator(self.cfg, self.proto, at.pubkey_type)
			ag = AddrGenerator(self.cfg, self.proto, at)
			return ag.to_addr(kg.gen_data(pk))

		compressed_types = set(self.proto.mmtypes) - {'L', 'E'}
		uncompressed_types = set(self.proto.mmtypes) & {'L', 'E'}

		def gen():
			for wif in key_list:
				pk = PrivKey(proto=self.proto, wif=wif)
				for t in (compressed_types if pk.compressed else uncompressed_types):
					yield (gen_addr(pk, t), pk)

		addrs4keys = dict(gen())

		for d in self.data:
			if d.addr in addrs4keys:
				d.sec = addrs4keys[d.addr]

	def list_missing(self, attr):
		return [d.addr for d in self.data if not getattr(d, attr)]

	@property
	def file(self):
		if not hasattr(self, '_file'):
			from . import addrfile
			self._file = getattr(addrfile, type(self).__name__.replace('List', 'File'))(self)
		return self._file

class KeyAddrList(AddrList):
	desc         = 'key-address'
	gen_desc     = 'key/address pair'
	gen_desc_pl  = 's'
	gen_keys     = True
	has_keys     = True
	chksum_rec_f = lambda foo, e: (str(e.idx), e.addr.views[e.addr.view_pref], e.sec.wif)

class ViewKeyAddrList(KeyAddrList):
	desc         = 'viewkey-address'
	gen_desc     = 'viewkey/address pair'
	chksum_rec_f = lambda foo, e: (str(e.idx), e.addr)

class KeyList(KeyAddrList):
	desc         = 'key'
	gen_desc     = 'key'
	gen_addrs    = False



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
swap.proto.thorchain.memo: THORChain swap protocol memo class
"""

from ....util import die, is_hex_str

from . import name as proto_name

from . import SwapAsset

class THORChainMemo:

	max_len = 250
	function = 'SWAP'

	function_abbrevs = {
		'SWAP': '=',
	}

	@classmethod
	def is_partial_memo(cls, bytes_data):
		import re
		ops = {
			'swap':     ('SWAP',     's',  '='),
			'add':      ('ADD',      'a',  r'\+'),
			'withdraw': ('WITHDRAW', 'wd', '-'),
			'loan':     (r'LOAN(\+|-)', r'\$(\+|-)'), # open/repay
			'pool':     (r'POOL(\+|-)',),
			'trade':    (r'TRADE(\+|-)',),
			'secure':   (r'SECURE(\+|-)',),
			'misc':     ('BOND', 'UNBOND', 'LEAVE', 'MIGRATE', 'NOOP', 'DONATE', 'RESERVE'),
		}
		pat = r'^(' + '|'.join('|'.join(pats) for pats in ops.values()) + r'):\S\S+'
		return bool(re.search(pat.encode(), bytes_data))

	@classmethod
	def parse(cls, s):
		"""
		All fields are validated, excluding address (cannot validate, since network is unknown)
		"""
		from collections import namedtuple
		from ....util import is_int

		def get_item(desc):
			try:
				return fields.pop(0)
			except IndexError:
				die('SwapMemoParseError', f'malformed {proto_name} memo (missing {desc} field)')

		def get_id(data, item, desc):
			if item in data:
				return item
			rev_data = {v:k for k,v in data.items()}
			if item in rev_data:
				return rev_data[item]
			die('SwapMemoParseError', f'{item!r}: unrecognized {proto_name} {desc} abbreviation')

		fields = str(s).split(':')

		if len(fields) < 4:
			die('SwapMemoParseError', 'memo must contain at least 4 comma-separated fields')

		function = get_id(cls.function_abbrevs, get_item('function'), 'function')

		asset = SwapAsset.init_from_memo(get_item('asset'))

		address = get_item('address')

		if asset.chain in SwapAsset.evm_chains:
			assert address.startswith('0x'), f'{address}: address does not start with ‘0x’'
			assert len(address) == 42, f'{address}: address has incorrect length ({len(address)} != 42)'
			address = address.removeprefix('0x')

		desc = 'trade_limit/stream_interval/stream_quantity'
		lsq = get_item(desc)

		try:
			limit, interval, quantity = lsq.split('/')
		except ValueError:
			die('SwapMemoParseError', f'malformed memo (failed to parse {desc} field) [{lsq}]')

		from . import ExpInt4
		try:
			limit_int = ExpInt4(limit)
		except Exception as e:
			die('SwapMemoParseError', str(e))

		for n in (interval, quantity):
			if not is_int(n):
				die('SwapMemoParseError', f'malformed memo (non-integer in {desc} field [{lsq}])')

		if fields:
			die('SwapMemoParseError', 'malformed memo (unrecognized extra data)')

		ret = namedtuple(
			'parsed_memo',
			['proto', 'function', 'asset', 'address', 'trade_limit', 'stream_interval', 'stream_quantity'])

		return ret(proto_name, function, asset, address, limit_int, int(interval), int(quantity))

	def __init__(self, swap_cfg, proto, asset, addr, *, trade_limit):

		from ....amt import UniAmt
		from ....addr import is_coin_addr

		assert trade_limit is None or isinstance(trade_limit, UniAmt), f'{type(trade_limit)} != {UniAmt}'
		assert is_coin_addr(proto, addr)

		assert asset.coin == proto.coin, f'{asset.coin} != {proto.coin}'
		assert asset.tokensym == getattr(proto, 'tokensym', None), (
			f'{asset.tokensym} != {getattr(proto, "tokensym", None)}')
		assert asset.direction == 'recv', f'{asset.direction} != ‘recv’'

		self.addr = addr.views[addr.view_pref]
		assert not ':' in self.addr # colon is record separator, so address mustn’t contain one

		if asset.chain in SwapAsset.evm_chains:
			assert len(self.addr) == 40, f'{self.addr}: address has incorrect length ({len(self.addr)} != 40)'
			assert is_hex_str(self.addr), f'{self.addr}: address is not a hexadecimal string'
			self.addr = '0x' + self.addr

		self.proto = proto
		self.asset = asset
		self.swap_cfg = swap_cfg
		self.trade_limit = trade_limit

	def __str__(self):
		from . import ExpInt4
		try:
			tl_enc = (
				0 if self.trade_limit is None else
				ExpInt4(self.trade_limit.to_unit('satoshi')).enc)
		except Exception as e:
			die('SwapMemoParseError', str(e))
		suf = '/'.join(str(n) for n in (
			tl_enc,
			self.swap_cfg.stream_interval,
			self.swap_cfg.stream_quantity))
		ret = ':'.join([
			self.function_abbrevs[self.function],
			self.asset.memo_asset_name,
			self.addr,
			suf])
		assert len(ret) <= self.max_len, f'{proto_name} memo exceeds maximum length of {self.max_len}'
		return ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.vm.tx.new: new transaction methods for VM chains
"""

from ....obj import MMGenTxID
from ....util import msg, is_int, is_hex_str, make_chksum_6, suf, die
from ....tw.ctl import TwCtl
from ....addr import is_mmgen_id, is_coin_addr

class New:
	desc = 'transaction'
	no_chg_msg = 'Warning: Transaction leaves account with zero balance'
	msg_insufficient_funds = 'Account balance insufficient to fund this transaction ({} {} needed)'

	# Instead of serializing tx data online as with BTC, put the data in a dict and serialize
	# offline before signing
	async def create_serialized(self, *, locktime=None):
		assert len(self.inputs) == 1, 'Transaction has more than one input!'
		o_num = len(self.outputs)
		o_ok = 0 if self.usr_contract_data else 1
		assert o_num == o_ok, f'Transaction has {o_num} output{suf(o_num)} (should have {o_ok})'
		await self.make_txobj()
		self.serialized = {k:v if v is None else str(v) for k, v in self.txobj.items() if k != 'token_to'}
		self.update_txid()

	def update_txid(self, data=None):
		import json
		assert not is_hex_str(self.serialized), (
			'update_txid() must be called only when self.serialized is not hex data')
		self.txid = MMGenTxID(make_chksum_6(json.dumps(data or self.serialized)).upper())

	async def process_cmdline_args(self, cmd_args, ad_f, ad_w):

		lc = len(cmd_args)

		if lc == 2 and self.is_swap:
			data_arg = cmd_args.pop()
			lc = 1
			assert data_arg.startswith('data:'), f'{data_arg}: invalid data arg (must start with "data:")'
			self.swap_memo = data_arg.removeprefix('data:')
			self.set_gas_with_data(self.swap_memo.encode())

		if lc == 0 and self.usr_contract_data and 'Token' not in self.name:
			return

		if lc != 1:
			die(1, f'{lc} output{suf(lc)} specified, but VM transactions must have exactly one')

		a = self.parse_cmdline_arg(self.proto, cmd_args[0], ad_f, ad_w)

		self.add_output(
			coinaddr = None if a.is_vault else a.addr,
			amt      = self.proto.coin_amt(a.amt or '0'),
			is_chg   = not a.amt,
			is_vault = a.is_vault)

		self.add_mmaddrs_to_outputs(ad_f, ad_w)

	def get_unspent_nums_from_user(self, unspent):
		from ....ui import line_input
		while True:
			reply = line_input(self.cfg, 'Enter an account to spend from: ').strip()
			if reply:
				if not is_int(reply):
					msg('Account number must be an integer')
				elif int(reply) < 1:
					msg('Account number must be >= 1')
				elif int(reply) > len(unspent):
					msg(f'Account number must be <= {len(unspent)}')
				else:
					return [int(reply)]

	def check_chg_addr_is_wallet_addr(self):
		pass

	def check_fee(self):
		if not self.disable_fee_check:
			assert self.usr_fee <= self.proto.max_tx_fee

	@property
	def total_gas(self):
		return self.gas

	def update_change_output(self, funds_left):
		if self.outputs and self.outputs[0].is_chg:
			self.update_output_amt(0, funds_left)

	async def get_input_addrs_from_inputs_opt(self):
		ret = []
		if self.cfg.inputs:
			data_root = (await TwCtl(self.cfg, self.proto)).data_root # must create new instance here
			errmsg = 'Address {!r} not in tracking wallet'
			for addr in self.cfg.inputs.split(','):
				if is_mmgen_id(self.proto, addr):
					for waddr in data_root:
						if data_root[waddr]['mmid'] == addr:
							ret.append(waddr)
							break
					else:
						die('UserAddressNotInWallet', errmsg.format(addr))
				elif is_coin_addr(self.proto, addr):
					if not addr in data_root:
						die('UserAddressNotInWallet', errmsg.format(addr))
					ret.append(addr)
				else:
					die(1, f'{addr!r}: not an MMGen ID or coin address')
		return ret

	def final_inputs_ok_msg(self, funds_left):
		chg = self.proto.coin_amt('0') if (self.outputs and self.outputs[0].is_chg) else funds_left
		return 'Transaction leaves {} {} in the sender’s account'.format(chg.hl(), self.proto.coin)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-seedjoin: Regenerate an MMGen deterministic wallet from seed shares
                created by 'mmgen-seedsplit'
"""

from .cfg import gc, Config
from .util import msg, msg_r, die
from .color import yellow
from .seed import Seed
from .seedsplit import SeedSplitIDString, MasterShareIdx, SeedShareMasterJoining
from .wallet import Wallet

opts_data = {
	'text': {
		'desc': """Regenerate an MMGen deterministic wallet from seed shares
                  created by 'mmgen-seedsplit'""",
		'usage': '[options] share1 share2 [...shareN]',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long (global) options
-d, --outdir=      d  Output file to directory 'd' instead of working dir
-e, --echo-passphrase Echo passphrases and other user input to screen
-i, --id-str=      s  ID String of split (required for master share join only)
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated).  NOTE: only the
                      first share may be in hidden incognito format!
-J, --hidden-incog-output-params=f,o  Write hidden incognito data to file
                      'f' at offset 'o' (comma-separated). File 'f' will be
                      created if necessary and filled with random data.
-o, --out-fmt=     f  Output to wallet format 'f' (see FMT CODES below)
-O, --old-incog-fmt   Specify old-format incognito input
-L, --label=       l  Specify a label 'l' for output wallet
-M, --master-share=i  Use a master share with index 'i' (min:{ms_min}, max:{ms_max})
-p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
                      for password hashing (default: '{gc.dfl_hash_preset}')
-z, --show-hash-presets Show information on available hash presets
-P, --passwd-file= f  Get wallet passphrase from file 'f'
-q, --quiet           Produce quieter output; suppress some warnings
-r, --usr-randchars=n Get 'n' characters of additional randomness from user
                      (min={cfg.min_urandchars}, max={cfg.max_urandchars}, default={cfg.usr_randchars})
-S, --stdout          Write wallet data to stdout instead of file
-v, --verbose         Produce more verbose output
""",
	'notes': """

COMMAND NOTES:

When joining with a master share, the master share must be listed first.
The remaining shares may be listed in any order.

The --id-str option is required only for master share joins.  For ordinary
joins it will be ignored.

For usage examples, see the help screen for the 'mmgen-seedsplit' command.

{n_pw}

{f}
"""
	},
	'code': {
		'options': lambda cfg, s: s.format(
			ms_min = MasterShareIdx.min_val,
			ms_max = MasterShareIdx.max_val,
			cfg    = cfg,
			gc     = gc,
		),
		'notes': lambda cfg, help_notes, s: s.format(
			f      = help_notes('fmt_codes'),
			n_pw   = help_notes('passwd'),
		)
	}
}

def print_shares_info():
	si = 0
	out = '\nComputed shares:\n'
	if cfg.master_share:
		fs = '{:3}: {}->{} ' + yellow('(master share #{}, split id ') + '{}' + yellow(', share count {})\n')
		out += fs.format(
				1,
				shares[0].sid,
				share1.sid,
				master_idx,
				id_str.hl2(encl='‘’'),
				len(shares))
		si = 1
	for n, s in enumerate(shares[si:], si+1):
		out += f'{n:3}: {s.sid}\n'
	cfg._util.qmsg(out)

cfg = Config(opts_data=opts_data)

if len(cfg._args) + bool(cfg.hidden_incog_input_params) < 2:
	cfg._usage()

if cfg.master_share:
	master_idx = MasterShareIdx(cfg.master_share)
	id_str = SeedSplitIDString(cfg.id_str or 'default')

if cfg.id_str and not cfg.master_share:
	die(1,'--id-str option meaningless in context of non-master-share join')

from .fileutil import check_infile
from .wallet import check_wallet_extension
for arg in cfg._args:
	check_wallet_extension(arg)
	check_infile(arg)

from .ui import do_license_msg
do_license_msg(cfg)

cfg._util.qmsg('Input files:\n  {}\n'.format('\n  '.join(cfg._args)))

shares = [Wallet(cfg).seed] if cfg.hidden_incog_input_params else []
shares += [Wallet(cfg, fn=fn).seed for fn in cfg._args]

if cfg.master_share:
	share1 = SeedShareMasterJoining(cfg, master_idx, shares[0], id_str, len(shares)).derived_seed
else:
	share1 = shares[0]

print_shares_info()

msg_r('Joining {n}-of-{n} XOR split...'.format(n=len(shares)))

seed_out = Seed.join_shares(cfg, [share1] + shares[1:])

msg(f'OK\nJoined Seed ID: {seed_out.sid.hl()}')

Wallet(cfg,seed=seed_out).write_to_file()



#!/usr/bin/env python3

"""
test.modtest_d.tx: TX unit tests for the MMGen suite
"""

import os

from mmgen.tx import CompletedTX, UnsignedTX
from mmgen.tx.file import MMGenTxFile
from mmgen.cfg import Config

from ..include.common import cfg, qmsg, vmsg, gr_uc

async def do_txfile_test(desc, fns, cfg=cfg, check=False):
	qmsg(f'\n  Testing CompletedTX initializer ({desc})')
	for fn in fns:
		qmsg(f'     parsing: {os.path.basename(fn)}')
		fpath = os.path.join('test', 'ref', fn)
		tx = await CompletedTX(cfg=cfg, filename=fpath, quiet_open=True)

		vmsg('\n' + tx.info.format())

		f = MMGenTxFile(tx)
		fn_gen = f.make_filename()

		if cfg.debug_utf8:
			fn_gen = fn_gen.replace('-α', '')

		assert fn_gen == os.path.basename(fn), f'{fn_gen} != {fn}'

		if check:
			import json
			from mmgen.tx.file import json_dumps
			from mmgen.util import make_chksum_6
			text = f.format()
			with open(fpath) as fh:
				text_chk = fh.read()
			data_chk = json.loads(text_chk)
			outputs = data_chk['MMGenTransaction']['outputs']
			for n, o in enumerate(outputs):
				outputs[n] = {k:v for k,v in o.items() if not (type(v) is bool and v is False)}
			data_chk['chksum'] = make_chksum_6(json_dumps(data_chk['MMGenTransaction']))
			text_chk_fixed = json_dumps(data_chk)
			assert text == text_chk_fixed, f'\nformatted text:\n{text}\n  !=\noriginal file:\n{text_chk_fixed}'

	qmsg('  OK')
	return True

class unit_tests:

	altcoin_deps = ('txfile_alt', 'txfile_alt_legacy')

	async def txfile(self, name, ut, desc='displaying transaction files (BTC)'):
		return await do_txfile_test(
			'Bitcoin',
			(
				'tx/7A8157[6.65227,34].rawtx',
				'tx/BB3FD2[7.57134314,123].sigtx',
				'tx/0A869F[1.23456,32].regtest.asubtx',
			),
			check = True
		)

	async def txfile_alt(self, name, ut, desc='displaying transaction files (LTC, BCH, ETH)'):
		return await do_txfile_test(
			'altcoins',
			(
				'tx/C09D73-LTC[981.73747,2000].testnet.rawtx',
				'tx/91060A-BCH[1.23456].regtest.arawtx',
				'tx/D850C6-MM1[43.21,50000].subtx', # token tx
			),
			# token resolved by tracking wallet under data_dir:
			cfg = Config({'data_dir': 'test/ref/data_dir'}),
			check = True
		)

	async def txfile_legacy(self, name, ut, desc='displaying transaction files (legacy format, BTC)'):
		return await do_txfile_test(
			'Bitcoin - legacy file format',
			(
				'0B8D5A[15.31789,14,tl=1320969600].rawtx',
				'542169[5.68152,34].sigtx',
				'0C7115[15.86255,14,tl=1320969600].testnet.rawtx',
				'25EFA3[2.34].testnet.rawtx',
			)
		)

	async def txfile_alt_legacy(self, name, ut, desc='displaying transaction files (legacy format, LTC, BCH, ETH)'):
		return await do_txfile_test(
			'altcoins - legacy file format',
			(
				'460D4D-BCH[10.19764,tl=1320969600].rawtx',
				'ethereum/5881D2-MM1[1.23456,50000].rawtx',
				'ethereum/6BDB25-MM1[1.23456,50000].testnet.rawtx',
				'ethereum/88FEFD-ETH[23.45495,40000].rawtx',
				'litecoin/A5A1E0-LTC[1454.64322,1453,tl=1320969600].testnet.rawtx',
				'litecoin/AF3CDF-LTC[620.76194,1453,tl=1320969600].rawtx',
			)
		)

	def errors(self, name, ut, desc='reading transaction files (error handling)'):
		async def bad1():
			await CompletedTX(cfg, filename='foo')
		def bad2():
			UnsignedTX(cfg, filename='foo')
		bad_data = (
			('forbidden positional args', 'TypeError', 'positional arguments', bad1),
			('forbidden positional args', 'TypeError', 'positional arguments', bad2),
		)
		ut.process_bad_data(bad_data)
		return True

	def op_return_data(self, name, ut, desc='OpReturnData class'):
		from mmgen.proto.btc.tx.op_return_data import OpReturnData
		vecs = [
			'data:=:ETH.ETH:0x86d526d6624AbC0178cF7296cD538Ecc080A95F1:0/1/0',
			'hexdata:3d3a4554482e4554483a30783836643532366436363234416243303137'
				'38634637323936634435333845636330383041393546313a302f312f30',
			'hexdata:00010203040506',
			'data:a\n',
			'data:a\tb',
			'data:' + gr_uc[:24],
		]

		assert OpReturnData(cfg._proto, vecs[0]) == OpReturnData(cfg._proto, vecs[1])

		for vec in vecs:
			d = OpReturnData(cfg._proto, vec)
			assert d == OpReturnData(cfg._proto, repr(d)) # repr() must return a valid initializer
			assert isinstance(d, bytes)
			assert isinstance(str(d), str)
			vmsg('-' * 80)
			vmsg(vec)
			vmsg(repr(d))
			vmsg(d.hl())
			vmsg(d.hl(add_label=True))

		bad_data = [
			'data:',
			'hexdata:',
			'data:' + ('x' * 81),
			'hexdata:' + ('deadbeef' * 20) + 'ee',
			'hex:0abc',
			'da:xyz',
			'hexdata:xyz',
			'hexdata:abcde',
			b'data:abc',
		]

		def bad(n):
			return lambda: OpReturnData(cfg._proto, bad_data[n])

		vmsg('-' * 80)
		vmsg('Testing error handling:')

		ut.process_bad_data((
			('bad1',    'AssertionError', 'not in range', bad(0)),
			('bad2',    'AssertionError', 'not in range', bad(1)),
			('bad3',    'AssertionError', 'not in range', bad(2)),
			('bad4',    'AssertionError', 'not in range', bad(3)),
			('bad5',    'ValueError',     'must start',   bad(4)),
			('bad6',    'ValueError',     'must start',   bad(5)),
			('bad7',    'AssertionError', 'not in hex',   bad(6)),
			('bad8',    'AssertionError', 'even',         bad(7)),
			('bad9',    'AssertionError', 'a string',     bad(8)),
		), pfx='')

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.online: online signed transaction class
"""

import sys, time, asyncio

from ..util import msg, Msg, gmsg, ymsg, make_timestr, die
from ..color import pink, yellow

from .signed import Signed, AutomountSigned

class OnlineSigned(Signed):

	@property
	def status(self):
		from . import _base_proto_subclass
		return _base_proto_subclass('Status', 'status', {'proto': self.proto})(self)

	def check_swap_expiry(self):
		from ..util2 import format_elapsed_hr
		expiry = self.swap_quote_expiry
		now = int(time.time())
		t_rem = expiry - now
		clr = yellow if t_rem < 0 else pink
		msg('Swap quote {a} {b} [{c}]'.format(
			a = clr('expired' if t_rem < 0 else 'expires'),
			b = clr(format_elapsed_hr(expiry, now=now, future_msg='from now')),
			c = make_timestr(expiry)))
		return t_rem >= 0

	def confirm_send(self, idxs):
		from ..ui import confirm_or_raise
		confirm_or_raise(
			cfg     = self.cfg,
			message = '' if self.cfg.quiet else 'Once this transaction is sent, there’s no taking it back!',
			action  = f'broadcast this transaction to the {self.proto.coin} {self.proto.network.upper()} network',
			expect  = 'YES' if self.cfg.quiet or self.cfg.yes else 'YES, I REALLY WANT TO DO THIS')
		msg('Sending transaction')
		if len(idxs) > 1 and getattr(self, 'coin_txid2', None) and self.is_swap:
			ymsg('Warning: two transactions (approval and router) will be broadcast to the network')

	async def post_send(self, asi):
		from . import SentTX
		tx2 = await SentTX(cfg=self.cfg, data=self.__dict__, automount=bool(asi))
		tx2.add_sent_timestamp()
		tx2.add_blockcount()
		tx2.file.write(
			outdir        = asi.txauto_dir if asi else None,
			ask_overwrite = False,
			ask_write     = False)

	async def send(self, cfg, asi):

		status_exitval = None
		sent_status = None
		all_ok = True
		idxs = ['', '2']

		if cfg.txhex_idx:
			if getattr(self, 'coin_txid2', None):
				if cfg.txhex_idx in ('1', '2'):
					idxs = ['' if cfg.txhex_idx == '1' else cfg.txhex_idx]
				else:
					die(1, f'{cfg.txhex_idx}: invalid parameter for --txhex-idx (must be 1 or 2)')
			else:
				die(1, 'Transaction has only one part, so --txhex-idx makes no sense')

		if not (cfg.status or cfg.receipt or cfg.dump_hex or cfg.test):
			self.confirm_send(idxs)

		for idx in idxs:
			if coin_txid := getattr(self, f'coin_txid{idx}', None):
				txhex = getattr(self, f'serialized{idx}')
				if cfg.status:
					cfg._util.qmsg(f'{self.proto.coin} txid: {coin_txid.hl()}')
					if cfg.verbose:
						await self.post_network_send(coin_txid)
					status_exitval = await self.status.display(idx=idx)
				elif cfg.receipt:
					if res := await self.get_receipt(coin_txid, receipt_only=True):
						import json
						Msg(json.dumps(res, indent=4))
					else:
						msg(f'Unable to get receipt for TX {coin_txid.hl()}')
				elif cfg.dump_hex:
					from ..fileutil import write_data_to_file
					write_data_to_file(
							cfg,
							cfg.dump_hex + idx,
							txhex + '\n',
							desc = 'serialized transaction hex data',
							ask_overwrite = False,
							ask_tty = False)
				elif cfg.tx_proxy:
					if idx != '' and not cfg.test_suite:
						await asyncio.sleep(2)
					from .tx_proxy import send_tx
					msg('{} TX: {}'.format('Testing' if cfg.test else 'Sending', coin_txid.hl()))
					if ret := send_tx(cfg, txhex):
						if ret != coin_txid:
							ymsg(f'Warning: txid mismatch (after sending) ({ret} != {coin_txid})')
						sent_status = 'confirm_post_send'
					if cfg.test:
						break
				elif cfg.test:
					if await self.test_sendable(txhex):
						gmsg('Transaction can be sent')
					else:
						ymsg('Transaction cannot be sent')
				else: # node send
					msg(f'Sending TX: {coin_txid.hl()}')
					if cfg.bogus_send:
						msg(f'BOGUS transaction NOT sent: {coin_txid.hl()}')
					else:
						if idx != '':
							await asyncio.sleep(1)
						ret = await self.send_with_node(txhex)
						msg(f'Transaction sent: {coin_txid.hl()}')
						assert ret == coin_txid, f'txid mismatch (after sending) ({ret} != {coin_txid})'
					sent_status = 'no_confirm_post_send'

				if cfg.wait and sent_status:
					res = await self.post_network_send(coin_txid)
					if all_ok:
						all_ok = res

		if not cfg.txhex_idx and sent_status and all_ok:
			from ..ui import keypress_confirm
			if sent_status == 'no_confirm_post_send' or not asi or keypress_confirm(
					cfg, 'Mark transaction as sent on removable device?'):
				await self.post_send(asi)

		if status_exitval is not None:
			if cfg.verbose:
				self.info.view_with_prompt('View transaction details?', pause=False)
			sys.exit(status_exitval)

class AutomountOnlineSigned(AutomountSigned, OnlineSigned):
	pass

class Sent(OnlineSigned):
	desc = 'sent transaction'
	ext = 'subtx'

class AutomountSent(AutomountOnlineSigned):
	desc = 'sent automount transaction'
	ext = 'asubtx'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.wallet: Wallet conversion tests for the cmdtest.py test suite
"""

import sys, os

from mmgen.util import msg, capfirst, get_extension
from mmgen.wallet import get_wallet_cls

from ..include.common import joinpath, VirtBlockDevice
from .include.common import ref_dir, ref_wallet_brainpass, ref_wallet_incog_offset, hincog_fn, hincog_bytes
from .base import CmdTestBase
from .shared import CmdTestShared

class CmdTestWalletConv(CmdTestBase, CmdTestShared):
	'wallet conversion to and from reference data'
	networks = ('btc',)
	tmpdir_nums = [11, 12, 13]
	sources = {
		'128': {
			'ref_wallet':     'FE3C6545-D782B529[128,1].mmdat',
			'ic_wallet':      'FE3C6545-E29303EA-5E229E30[128,1].mmincog',
			'ic_wallet_hex':  'FE3C6545-BC4BE3F2-32586837[128,1].mmincox',

			'hic_wallet':      'FE3C6545-161E495F-BEB7548E[128,1].incog-offset123',
			'hic_wallet_old':  'FE3C6545-161E495F-9860A85B[128,1].incog-old.offset123',
		},
		'192': {
			'ref_wallet':     '1378FC64-6F0F9BB4[192,1].mmdat',
			'ic_wallet':      '1378FC64-2907DE97-F980D21F[192,1].mmincog',
			'ic_wallet_hex':  '1378FC64-4DCB5174-872806A7[192,1].mmincox',

			'hic_wallet':     '1378FC64-B55E9958-77256FC1[192,1].incog.offset123',
			'hic_wallet_old': '1378FC64-B55E9958-D85FF20C[192,1].incog-old.offset123',
		},
		'256': {
			'ref_wallet':     '98831F3A-27F2BF93[256,1].mmdat',
			'ic_wallet':      '98831F3A-5482381C-18460FB1[256,1].mmincog',
			'ic_wallet_hex':  '98831F3A-1630A9F2-870376A9[256,1].mmincox',

			'hic_wallet':      '98831F3A-F59B07A0-559CEF19[256,1].incog.offset123',
			'hic_wallet_old':  '98831F3A-F59B07A0-848535F3[256,1].incog-old.offset123',

		},
	}
	cmd_group = (
		# reading
		('ref_wallet_conv',            'conversion of saved reference wallet'),
		('ref_mn_conv',                'conversion of saved MMGen native mnemonic'),
		('ref_bip39_conv',             'conversion of saved BIP39 mnemonic'),
		('ref_seed_conv',              'conversion of saved seed file'),
		('ref_hex_conv',               'conversion of saved MMGen hexadecimal seed file'),
		('ref_plainhex_conv',          'conversion of saved plain hexadecimal seed file'),
		('ref_dieroll_conv',           'conversion of saved dieroll (b6d) seed file'),
		('ref_brain_conv',             'conversion of ref brainwallet'),
		('ref_incog_conv',             'conversion of saved incog wallet'),
		('ref_incox_conv',             'conversion of saved hex incog wallet'),
		('ref_hincog_conv',            'conversion of saved hidden incog wallet'),
		('ref_hincog_conv_old',        'conversion of saved hidden incog wallet (old format)'),
		# writing
		('ref_wallet_conv_out',        'ref seed conversion to wallet'),
		('ref_mn_conv_out',            'ref seed conversion to MMGen native mnemonic'),
		('ref_bip39_conv_out',         'ref seed conversion to BIP39 mnemonic'),
		('ref_hex_conv_out',           'ref seed conversion to MMGen hex seed'),
		('ref_plainhex_conv_out',      'ref seed conversion to plain hex seed'),
		('ref_dieroll_conv_out',       'ref seed conversion to dieroll (b6d) seed'),
		('ref_seed_conv_out',          'ref seed conversion to seed'),
		('ref_incog_conv_out',         'ref seed conversion to incog data'),
		('ref_incox_conv_out',         'ref seed conversion to hex incog data'),
		('ref_hincog_conv_out',        'ref seed conversion to hidden incog data'),
		('ref_hincog_blkdev_conv_out', 'ref seed conversion to hidden incog data on block device')
	)

	def __init__(self, cfg, trunner, cfgs, spawn):
		for k, _ in self.cmd_group:
			for n in (1, 2, 3):
				setattr(self, f'{k}_{n}', getattr(self, k))
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)

	def ref_wallet_conv(self):
		wf = joinpath(ref_dir, self.sources[str(self.seed_len)]['ref_wallet'])
		return self.walletconv_in(wf)

	def ref_mn_conv(self, ext='mmwords'):
		wf = joinpath(ref_dir, self.seed_id+'.'+ext)
		return self.walletconv_in(wf)

	def ref_bip39_conv(self):
		return self.ref_mn_conv(ext='bip39')
	def ref_seed_conv(self):
		return self.ref_mn_conv(ext='mmseed')
	def ref_hex_conv(self):
		return self.ref_mn_conv(ext='mmhex')
	def ref_plainhex_conv(self):
		return self.ref_mn_conv(ext='hex')
	def ref_dieroll_conv(self):
		return self.ref_mn_conv(ext='b6d')

	def ref_brain_conv(self):
		uopts = ['-i', 'bw', '-p', '1', '-l', str(self.seed_len)]
		return self.walletconv_in(None, uopts, icls=get_wallet_cls('brain'))

	def ref_incog_conv(self, wfk='ic_wallet', in_fmt='i'):
		uopts = ['-i', in_fmt, '-p', '1', '-l', str(self.seed_len)]
		wf = joinpath(ref_dir, self.sources[str(self.seed_len)][wfk])
		return self.walletconv_in(wf, uopts)

	def ref_incox_conv(self):
		return self.ref_incog_conv(in_fmt='xi', wfk='ic_wallet_hex')

	def ref_hincog_conv(self, wfk='hic_wallet', add_uopts=[]):
		ic_f = joinpath(ref_dir, self.sources[str(self.seed_len)][wfk])
		uopts = ['-i', 'hi', '-p', '1', '-l', str(self.seed_len)] + add_uopts
		hi_opt = ['-H', f'{ic_f},{ref_wallet_incog_offset}']
		return self.walletconv_in(
			None,
			uopts + hi_opt,
			icls = get_wallet_cls('incog_hidden'))

	def ref_hincog_conv_old(self):
		return self.ref_hincog_conv(wfk='hic_wallet_old', add_uopts=['-O'])

	def ref_wallet_conv_out(self):
		return self.walletconv_out('w')
	def ref_mn_conv_out(self):
		return self.walletconv_out('mn')
	def ref_bip39_conv_out(self):
		return self.walletconv_out('bip39')
	def ref_seed_conv_out(self):
		return self.walletconv_out('seed')
	def ref_hex_conv_out(self):
		return self.walletconv_out('hexseed')
	def ref_plainhex_conv_out(self):
		return self.walletconv_out('hex')
	def ref_dieroll_conv_out(self):
		return self.walletconv_out('dieroll')
	def ref_incog_conv_out(self):
		return self.walletconv_out('i')
	def ref_incox_conv_out(self):
		return self.walletconv_out('xi')

	def ref_hincog_conv_out(self, ic_f=None):
		if not ic_f:
			ic_f = joinpath(self.tmpdir, hincog_fn)
		hi_parms = f'{ic_f},{ref_wallet_incog_offset}'
		sl_parm = '-l' + str(self.seed_len)
		return self.walletconv_out(
			'hi',
			uopts     = ['-J', hi_parms, sl_parm],
			uopts_chk = ['-H', hi_parms, sl_parm])

	def ref_hincog_blkdev_conv_out(self):

		if self.skip_for_win('no loop device'):
			return 'skip'

		b = VirtBlockDevice(os.path.join(self.tmpdir, 'hincog_blkdev_img'), '1K')
		b.create()
		b.attach(dev_mode='0666' if sys.platform == 'linux' else None)
		self.ref_hincog_conv_out(ic_f=b.dev)
		b.detach()

		return 'ok'

	# wallet conversion tests
	def walletconv_in(self, infile, uopts=[], icls=None):
		ocls = get_wallet_cls('words')
		opts = ['-d', self.tmpdir, '-o', ocls.fmt_codes[0], self.usr_rand_arg]
		if_arg = [infile] if infile else []
		d = '(convert)'
		t = self.spawn('mmgen-walletconv', opts+uopts+if_arg, extra_desc=d)
		t.license()
		icls = icls or get_wallet_cls(ext=get_extension(infile))
		if icls.type == 'brain':
			t.expect('Enter brainwallet: ', ref_wallet_brainpass+'\n')
		if icls.enc and icls.type != 'brain':
			t.passphrase(icls.desc, self.wpasswd)
			if self.test_name[:19] == 'ref_hincog_conv_old':
				t.expect('Is the Seed ID correct? (Y/n): ', '\n')
			else:
				t.expect(['Passphrase is OK', ' are correct'])
		wf = t.written_to_file(capfirst(ocls.desc))
		t.p.wait()
		# back check of result
		msg('' if self.cfg.profile else ' OK')
		return self.walletchk(
			wf,
			extra_desc = '(check)',
			sid        = self.seed_id)

	def walletconv_out(self, out_fmt='w', uopts=[], uopts_chk=[]):
		wcls = get_wallet_cls(fmt_code=out_fmt)
		opts = ['-d', self.tmpdir, '-p1', '-o', out_fmt] + uopts
		infile = joinpath(ref_dir, self.seed_id+'.mmwords')
		t = self.spawn('mmgen-walletconv', [self.usr_rand_arg]+opts+[infile], extra_desc='(convert)')

		add_args = [f'-l{self.seed_len}']
		t.license()
		if wcls.enc and wcls.type != 'brain':
			t.passphrase_new('new '+wcls.desc, self.wpasswd)
			t.usr_rand(self.usr_rand_chars)
		if wcls.type.startswith('incog'):
			for _ in range(3):
				t.expect('Encrypting random data from your operating system with ephemeral key')
		if wcls.type == 'incog_hidden':
			t.hincog_create(hincog_bytes)
		if out_fmt == 'w':
			t.label()
		wf = t.written_to_file(capfirst(wcls.desc))

		if wcls.type == 'incog_hidden':
			add_args += uopts_chk
			wf = None
		msg('' if self.cfg.profile else ' OK')
		return self.walletchk(
			wf,
			wcls       = wcls,
			extra_desc = '(check)',
			sid        = self.seed_id,
			add_args   = add_args)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.signed: Ethereum signed transaction class
"""

from ....tx import signed as TxBase
from ....obj import CoinTxID, ETHNonce, HexStr
from ....addr import CoinAddr, ContractAddr
from .completed import Completed, TokenCompleted

class Signed(Completed, TxBase.Signed):

	desc = 'signed transaction'

	def parse_txfile_serialized_data(self):
		from .transaction import Transaction
		from .. import rlp
		etx = rlp.decode(bytes.fromhex(self.serialized), Transaction)
		d = etx.to_dict() # ==> hex values have '0x' prefix, 0 is '0x'
		for k in ('sender', 'to', 'data'):
			if k in d:
				d[k] = d[k].replace('0x', '', 1)
		o = {
			'from':     CoinAddr(self.proto, d['sender']),
			# NB: for token, 'to' is token address
			'to':       CoinAddr(self.proto, d['to']) if d['to'] else None,
			'amt':      self.proto.coin_amt(d['value'], from_unit='wei'),
			'gasPrice': self.proto.coin_amt(d['gasprice'], from_unit='wei'),
			'startGas': d['startgas'],
			'nonce':    ETHNonce(d['nonce']),
			'data':     HexStr(d['data']) }
		if o['data'] and not o['to']: # token- or contract-creating transaction
			# NB: could be a non-token contract address:
			o['token_addr'] = ContractAddr(self.proto, etx.creates.hex())
			self.disable_fee_check = True
		txid = CoinTxID(etx.hash.hex())
		assert txid == self.coin_txid, "txid in tx.serialized doesn't match value in MMGen transaction file"
		self.txobj = o
		return d # 'token_addr', 'decimals' required by Token subclass

class TokenSigned(TokenCompleted, Signed):
	desc = 'signed transaction'

	def parse_txfile_serialized_data(self):
		raise NotImplementedError(
			'Signed token transaction files cannot be parsed offline, because tracking wallet is required!')

class AutomountSigned(TxBase.AutomountSigned, Signed):
	pass

class TokenAutomountSigned(TxBase.AutomountSigned, TokenSigned):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tw.ctl: Tracking wallet control class for the MMGen suite
"""

from collections import namedtuple

from ..util import msg, msg_r, ymsg, suf, die
from ..base_obj import AsyncInit
from ..objmethods import MMGenObject
from ..obj import TwComment, get_obj
from ..addr import CoinAddr, is_mmgen_id, is_coin_addr
from ..rpc import rpc_init
from .shared import TwMMGenID, TwLabel

twmmid_addr_pair = namedtuple('addr_info', ['twmmid', 'coinaddr'])
label_addr_pair = namedtuple('label_addr_pair', ['label', 'coinaddr'])

# decorator for TwCtl
def write_mode(orig_func):
	def f(self, *args, **kwargs):
		if self.mode != 'w':
			die(1, '{} opened in read-only mode: cannot execute method {}()'.format(
				type(self).__name__,
				locals()['orig_func'].__name__
			))
		return orig_func(self, *args, **kwargs)
	return f

class TwCtl(MMGenObject, metaclass=AsyncInit):

	caps = ('rescan', 'batch')
	data_key = 'addresses'
	importing = False
	use_cached_balances = False

	def __new__(cls, cfg, proto, *args, **kwargs):
		return MMGenObject.__new__(
			proto.base_proto_subclass(cls, 'tw.ctl', is_token=kwargs.get('token_addr')))

	async def __init__(
			self,
			cfg,
			proto,
			*,
			mode              = 'r',
			token_addr        = None,
			no_rpc            = False,
			rpc_ignore_wallet = False):

		assert mode in ('r', 'w', 'i'), f"{mode!r}: wallet mode must be 'r', 'w' or 'i'"
		if mode == 'i':
			self.importing = True
			mode = 'w'

		self.cfg = cfg
		self.proto = proto
		self.mode = mode
		self.desc = self.base_desc = f'{self.proto.name} tracking wallet'

		if not no_rpc:
			self.rpc = await rpc_init(cfg, proto, ignore_wallet=rpc_ignore_wallet)

	async def resolve_address(self, addrspec):

		twmmid, coinaddr = (None, None)

		pairs = await self.get_label_addr_pairs()

		if is_coin_addr(self.proto, addrspec):
			coinaddr = get_obj(CoinAddr, proto=self.proto, addr=addrspec)
			pair_data = [e for e in pairs if e.coinaddr == coinaddr]
		elif is_mmgen_id(self.proto, addrspec):
			twmmid = TwMMGenID(self.proto, addrspec)
			pair_data = [e for e in pairs if e.label.mmid == twmmid]
		else:
			msg(f'{addrspec!r}: invalid address for this network')
			return None

		if not pair_data:
			msg('{a} address {b!r} not found in tracking wallet'.format(
				a = 'MMGen' if twmmid else 'Coin',
				b = twmmid or coinaddr))
			return None

		return twmmid_addr_pair(
			twmmid or pair_data[0].label.mmid,
			coinaddr or pair_data[0].coinaddr)

	# returns on failure
	@write_mode
	async def set_comment(
			self,
			addrspec,
			comment      = '',
			*,
			trusted_pair = None,
			silent       = False):

		res = twmmid_addr_pair(*trusted_pair) if trusted_pair else await self.resolve_address(addrspec)

		if not res:
			return False

		comment = get_obj(TwComment, s=comment)

		if comment is False:
			return False

		lbl = get_obj(
			TwLabel,
			proto = self.proto,
			text = res.twmmid + (' ' + comment if comment else ''))

		if lbl is False:
			return False

		if await self.set_label(res.coinaddr, lbl):
			if not silent:
				desc = '{t} address {a} in tracking wallet'.format(
					t = res.twmmid.type.replace('mmgen', 'MMGen'),
					a = res.twmmid.addr.hl() if res.twmmid.type == 'mmgen' else
						res.twmmid.addr.hl(res.twmmid.addr.view_pref))
				msg(
					'Added label {} to {}'.format(comment.hl2(encl='‘’'), desc) if comment else
					'Removed label from {}'.format(desc))
			return True
		else:
			if not silent:
				msg('Label could not be {}'.format('added' if comment else 'removed'))
			return False

	@write_mode
	async def remove_comment(self, mmaddr):
		await self.set_comment(mmaddr, '')

	def check_import_mmid(self, addr, old_mmid, new_mmid):
		'returns True if mmid needs update, None otherwise'
		if new_mmid != old_mmid:
			if old_mmid.endswith(':' + addr):
				ymsg(f'Warning: address {new_mmid} was previously imported as non-MMGen!')
				return True
			else:
				fs = (
					'attempting to import MMGen address {a!r} ({b}) as non-MMGen!'
						if new_mmid.endswith(':' + addr) else
					'imported MMGen ID {b!r} does not match tracking wallet MMGen ID {a!r}!')
				die(2, fs.format(a=old_mmid, b=new_mmid))

	async def import_address_common(self, data, *, batch=False, gather=False):

		async def do_import(address, comment, message):
			try:
				res = await self.import_address(address, label=comment)
				self.cfg._util.qmsg(message)
				return res
			except Exception as e:
				die(2, f'\nImport of address {address!r} failed: {e.args[0]!r}')

		_d = namedtuple('formatted_import_data', data[0]._fields + ('mmid_disp',))
		pfx = self.proto.base_coin.lower() + ':'
		fdata = [_d(*d, 'non-MMGen' if d.twmmid.startswith(pfx) else d.twmmid) for d in data]

		fs = '{:%s}: {:%s} {:%s} - OK' % (
			len(str(len(fdata))) * 2 + 1,
			max(len(d.addr) for d in fdata),
			max(len(d.mmid_disp) for d in fdata) + 2
		)

		nAddrs = len(data)
		out = [( # create list, not generator, so we know data is valid before starting import
				CoinAddr(self.proto, d.addr),
				TwLabel(self.proto, d.twmmid + (f' {d.comment}' if d.comment else '')),
				fs.format(f'{n}/{nAddrs}', d.addr, f'({d.mmid_disp})')
			) for n, d in enumerate(fdata, 1)]

		if batch:
			msg_r(f'Batch importing {len(out)} address{suf(data, "es")}...')
			ret = await self.batch_import_address((a, b, False) for a, b, c in out)
			msg(f'done\n{len(ret)} addresses imported')
		else:
			if gather: # this seems to provide little performance benefit
				import asyncio
				await asyncio.gather(*(do_import(*d) for d in out))
			else:
				for d in out:
					await do_import(*d)
			msg('Address import completed OK')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.misc: miscellaneous functions for Bitcoin base protocol
"""

from ...util import msg, msg_r

async def scantxoutset(cfg, rpc, descriptor_list):

	import asyncio

	async def do_scan():
		return await rpc.call(
			'scantxoutset',
			'start',
			descriptor_list,
			timeout = 720) # call may take several minutes to complete

	async def do_status():

		CR = '\n' if cfg.test_suite else '\r'
		sleep_secs = 0.1 if cfg.test_suite else 2
		m = f'{CR}Scanning UTXO set: '
		msg_r(m + '0% completed ')

		while True:
			await asyncio.sleep(sleep_secs)
			res = await rpc.call('scantxoutset', 'status')
			if res:
				msg_r(m + f'{res["progress"]}% completed ')
			if task1.done():
				msg(m + '100% completed')
				return

	res = await rpc.call('scantxoutset', 'status')
	if res and res.get('progress'):
		msg_r('Aborting scan in progress...')
		await rpc.call('scantxoutset', 'abort')
		await asyncio.sleep(1)
		msg('done')

	if rpc.backend.name == 'aiohttp':
		task1 = asyncio.create_task(do_scan())
		task2 = asyncio.create_task(do_status())
		ret = await task1
		await task2
	else:
		msg_r('Scanning UTXO set, this could take several minutes...')
		ret = await do_scan()
		msg('done')

	return ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.ctl: Bitcoin base protocol tracking wallet control class
"""

from ....tw.ctl import TwCtl, write_mode
from ....util import msg, msg_r, rmsg, die, suf, fmt_list

class BitcoinTwCtl(TwCtl):

	async def rpc_get_balance(self, addr, block='latest'):
		raise NotImplementedError('not implemented')

	# TODO: do check with check_import_mmid()
	@write_mode
	async def import_address(self, addr, *, label, rescan=False):
		if (await self.rpc.walletinfo).get('descriptors'):
			return await self.batch_import_address([(addr, label, rescan)])
		else:
			return await self.rpc.call('importaddress', addr, label, rescan)

	@write_mode
	async def batch_import_address(self, arg_list):
		if (await self.rpc.walletinfo).get('descriptors'):
			from ....contrib.descriptors import descsum_create
			return await self.rpc.call(
				'importdescriptors',
				[{
					'desc': descsum_create(f'addr({addr})'),
					'label': label,
					'timestamp': 0 if rescan else 'now',
				} for addr, label, rescan in arg_list])
		else:
			return await self.rpc.batch_call('importaddress', arg_list)

	@write_mode
	async def remove_address(self, addr):
		raise NotImplementedError(f'address removal not implemented for coin {self.proto.coin}')

	@write_mode
	async def set_label(self, coinaddr, lbl):
		args = self.rpc.daemon.set_comment_args(self.rpc, coinaddr, lbl)
		try:
			await self.rpc.call(*args)
			return True
		except Exception as e:
			rmsg(e.args[0])
			return False

	@write_mode
	async def rescan_blockchain(self, start, stop):

		start = start or 0
		endless = stop is None
		CR = '\n' if self.cfg.test_suite else '\r'

		if not (start >= 0 and (stop if stop is not None else start) >= start):
			die(1, f'{start} {stop}: invalid range')

		async def do_scan(chunks, tip):
			res = None
			for a, b in chunks:
				msg_r(f'{CR}Scanning blocks {a}-{b} ')
				res = await self.rpc.call('rescanblockchain', a, b, timeout=7200)
				if res['start_height'] != a or res['stop_height'] != b:
					die(1, f'\nAn error occurred in block range {a}-{b}')
			msg('')
			return b if res else tip

		def gen_chunks(start, stop, tip):
			n = start
			if endless:
				stop = tip
			elif stop > tip:
				die(1, f'{stop}: stop value is higher than chain tip')

			while n <= stop:
				yield (n, min(n+99, stop))
				n += 100

		last_block = await do_scan(gen_chunks(start, stop, self.rpc.blockcount), self.rpc.blockcount)

		if endless:
			tip = await self.rpc.call('getblockcount')
			while last_block < tip:
				last_block = await do_scan(gen_chunks(last_block+1, tip, tip), tip)
				tip = await self.rpc.call('getblockcount')

		msg('Done')

	@write_mode
	async def rescan_address(self, addrspec):
		res = await self.resolve_address(addrspec)
		if not res:
			return False
		return await self.rescan_addresses([res.coinaddr])

	@write_mode
	async def rescan_addresses(self, coin_addrs):

		from ..misc import scantxoutset
		res = await scantxoutset(self.cfg, self.rpc, [f'addr({addr})' for addr in coin_addrs])

		if not res['success']:
			msg('UTXO scanning failed or was interrupted')
			return False
		elif res['unspents']:
			blocks = sorted({i['height'] for i in res['unspents']})
			msg('Found {} unspent output{} in {} block{}'.format(
				len(res['unspents']),
				suf(res['unspents']),
				len(blocks),
				suf(blocks)))
			self.cfg._util.vmsg(f'Blocks to rescan: {fmt_list(blocks, fmt="bare")}')
			CR = '\n' if self.cfg.test_suite else '\r'
			for n, block in enumerate(blocks):
				msg_r(f'{CR}Rescanning block: {block} ({n+1}/{len(blocks)})')
				# httplib seems to require fresh connection here, so specify timeout
				await self.rpc.call('rescanblockchain', block, block, timeout=60)
			msg(f'\nAddress balance{suf(coin_addrs)} updated successfully')
			return True
		else:
			msg('Address has no balance' if len(coin_addrs) == 1 else
				'Addresses have no balances')
			return True

	async def get_label_addr_pairs(self):
		from .rpc import TwRPC
		return await TwRPC(proto=self.proto, rpc=self.rpc, twctl=self).get_label_addr_pairs()



#!/usr/bin/env python3
# this script is used both for interactive and automated testing

import sys, os
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(os.path.dirname(sys.argv[0])))))
sys.path[0] = os.curdir

from mmgen.cfg import Config
from mmgen.util import msg

opts_data = {
	'text': {
		'desc': '',
		'usage':'',
		'options':'-e, --echo-passphrase foo',
	}
}

cfg = Config(opts_data=opts_data)

cmd_args = cfg._args

cmd = cmd_args[0]

if cmd == 'passphrase':
	from mmgen.ui import get_words_from_user
	pw = get_words_from_user(
		cfg,
		('Enter passphrase: ', 'Enter passphrase (echoed): ')[bool(cfg.echo_passphrase)] )
	msg('Entered: {}'.format(' '.join(pw)))
elif cmd in ('get_char', 'line_input'):
	from mmgen.term import get_char
	from mmgen.ui import line_input
	from ast import literal_eval
	func_args = literal_eval(cmd_args[1])
	msg(f'\n  term: {get_char.__self__.__name__}')
	msg(f'  cfg.hold_protect_disable: {cfg.hold_protect_disable}')
	if cmd == 'line_input':
		func_args.update({'cfg':cfg})
	msg('  Calling {name}({args})'.format(
		name = cmd,
		args = ', '.join(f'{k}={v!r}' for k, v in func_args.items())
		))
	ret = locals()[cmd](**func_args)
	msg(f'  ==> {ret!r}')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
bip39.py - Data and routines for BIP39 mnemonic seed phrases
"""

from hashlib import sha256

from .baseconv import baseconv
from .util import is_hex_str, die

def is_bip39_mnemonic(s):
	return bool(bip39().tohex(s.split()))

# implements a subset of the baseconv API
class bip39(baseconv):

	desc            = baseconv.dt('BIP39 mnemonic', 'BIP39 mnemonic seed phrase')
	wl_chksum       = 'f18b9a84'
	seedlen_map     = {16:12, 24:18, 32:24}
	seedlen_map_rev = {12:16, 18:24, 24:32}

	from collections import namedtuple
	bc = namedtuple('bip39_constants', ['chk_len', 'mn_len'])
	#    ENT   CS  MS
	constants = {
		128: bc(4, 12),
		160: bc(5, 15),
		192: bc(6, 18),
		224: bc(7, 21),
		256: bc(8, 24),
	}

	def __init__(self, wl_id='bip39'):
		assert wl_id == 'bip39', "initialize with 'bip39' for compatibility with baseconv API"
		from .wordlist.bip39 import words
		self.digits = words
		self.wl_id = 'bip39'

	@classmethod
	def nwords2seedlen(cls, nwords, /, *, in_bytes=False, in_hex=False):
		for k, v in cls.constants.items():
			if v.mn_len == nwords:
				return k//8 if in_bytes else k//4 if in_hex else k
		die('MnemonicError', f'{nwords!r}: invalid word length for BIP39 mnemonic')

	@classmethod
	def seedlen2nwords(cls, seed_len, /, *, in_bytes=False, in_hex=False):
		seed_bits = seed_len * 8 if in_bytes else seed_len * 4 if in_hex else seed_len
		try:
			return cls.constants[seed_bits].mn_len
		except Exception as e:
			raise ValueError(f'{seed_bits!r}: invalid seed length for BIP39 mnemonic') from e

	def tohex(self, words_arg, /, *, pad=None):
		return self.tobytes(words_arg, pad=pad).hex()

	def tobytes(self, words_arg, /, *, pad=None):
		assert isinstance(words_arg, (list, tuple)), 'words_arg must be list or tuple'
		assert pad in (None, 'seed'), f"{pad}: invalid 'pad' argument (must be None or 'seed')"

		wl = self.digits

		for n, w in enumerate(words_arg):
			if w not in wl:
				die('MnemonicError', f'word #{n+1} is not in the BIP39 word list')

		res = ''.join(f'{wl.index(w):011b}' for w in words_arg)

		for k, v in self.constants.items():
			if len(words_arg) == v.mn_len:
				bitlen = k
				break
		else:
			die('MnemonicError', f'{len(words_arg)}: invalid BIP39 seed phrase length')

		seed_bin = res[:bitlen]
		chk_bin = res[bitlen:]

		seed_hex = f'{int(seed_bin, 2):0{bitlen//4}x}'
		seed_bytes = bytes.fromhex(seed_hex)

		chk_len = self.constants[bitlen].chk_len
		chk_hex_chk = sha256(seed_bytes).hexdigest()
		chk_bin_chk = f'{int(chk_hex_chk, 16):0256b}'[:chk_len]

		if chk_bin != chk_bin_chk:
			die('MnemonicError', f'invalid BIP39 seed phrase checksum ({chk_bin} != {chk_bin_chk})')

		return seed_bytes

	def fromhex(self, hexstr, /, *, pad=None, tostr=False):
		assert is_hex_str(hexstr), 'seed data not a hexadecimal string'
		return self.frombytes(bytes.fromhex(hexstr), pad=pad, tostr=tostr)

	def frombytes(self, seed_bytes, /, *, pad=None, tostr=False):
		assert tostr is False, "'tostr' must be False for 'bip39'"
		assert pad in (None, 'seed'), f"{pad}: invalid 'pad' argument (must be None or 'seed')"

		wl = self.digits
		bitlen = len(seed_bytes) * 8

		assert bitlen in self.constants, f'{bitlen}: invalid seed bit length'
		c = self.constants[bitlen]

		chk_hex = sha256(seed_bytes).hexdigest()

		seed_bin = f'{int(seed_bytes.hex(), 16):0{bitlen}b}'
		chk_bin  = f'{int(chk_hex, 16):0256b}'

		res = seed_bin + chk_bin

		return tuple(wl[int(res[i*11:(i+1)*11], 2)] for i in range(c.mn_len))

	def generate_seed(self, words_arg, /, *, passwd=''):

		self.tohex(words_arg) # validate

		from cryptography.hazmat.primitives import hashes
		from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
		return PBKDF2HMAC(
			algorithm  = hashes.SHA512(),
			length     = 64,
			salt       = b'mnemonic' + passwd.encode(),
			iterations = 2048
		).derive(' '.join(words_arg).encode())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.new: new transaction class
"""

from collections import namedtuple

from .base import Base
from ..cfg import gc
from ..color import pink, yellow
from ..obj import get_obj, MMGenList
from ..util import msg, fmt, die, suf, remove_dups, get_extension
from ..addr import (
	is_mmgen_id,
	MMGenAddrType,
	MMGenID,
	CoinAddr,
	is_mmgen_addrtype,
	is_coin_addr,
	is_addrlist_id
)

def mmaddr2coinaddr(cfg, mmaddr, ad_w, ad_f, proto):

	def wmsg(k):
		messages = {
			'addr_in_addrfile_only': f"""
				Warning: output address {mmaddr} is not in the tracking wallet, which
				means its balance will not be tracked.  You're strongly advised to import
				the address into your tracking wallet before broadcasting this transaction.
			""",
			'addr_not_found': f"""
				No data for {gc.proj_name} address {mmaddr} could be found in either the
				tracking wallet or the supplied address file.  Please import this address
				into your tracking wallet, or supply an address file on the command line.
			""",
			'addr_not_found_no_addrfile': f"""
				No data for {gc.proj_name} address {mmaddr} could be found in the tracking
				wallet.  Please import this address into your tracking wallet or supply an
				address file for it on the command line.
			"""
		}
		return '\n' + fmt(messages[k], indent='  ')

	# assume mmaddr has already been checked
	coin_addr = ad_w.mmaddr2coinaddr(mmaddr)

	if not coin_addr:
		if ad_f:
			coin_addr = ad_f.mmaddr2coinaddr(mmaddr)
			if coin_addr:
				msg(wmsg('addr_in_addrfile_only'))
				from ..ui import keypress_confirm
				if not (cfg.yes or keypress_confirm(cfg, 'Continue anyway?')):
					import sys
					sys.exit(1)
			else:
				die(2, wmsg('addr_not_found'))
		else:
			die(2, wmsg('addr_not_found_no_addrfile'))

	return CoinAddr(proto, coin_addr)

def parse_fee_spec(proto, fee_arg):
	import re
	units = {u[0]:u for u in proto.coin_amt.units}
	pat = re.compile(r'((?:[1-9][0-9]*)|(?:[0-9]+\.[0-9]+))({})'.format('|'.join(units)))
	if m := pat.match(fee_arg):
		return namedtuple('parsed_fee_spec', ['amt', 'unit'])(m[1], units[m[2]])

class New(Base):

	fee_is_approximate = False
	msg_wallet_low_coin = 'Wallet has insufficient funds for this transaction ({} {} needed)'
	msg_no_change_output = """
		ERROR: No change address specified.  If you wish to create a transaction with
		only one output, specify a single output address with no {} amount
	"""
	chg_autoselected = False
	_funds_available = namedtuple('funds_available', ['is_positive', 'amt'])
	_net_fee = namedtuple('network_fee_estimate', ['fee', 'type'])

	def warn_insufficient_funds(self, amt, coin):
		msg(self.msg_insufficient_funds.format(amt.hl(), coin))

	def update_output_amt(self, idx, amt):
		o = self.outputs[idx]._asdict()
		o['amt'] = amt
		self.outputs[idx] = self.Output(self.proto, **o)

	def add_mmaddrs_to_outputs(self, ad_f, ad_w):
		a = [e.addr for e in self.outputs]
		d = ad_w.make_reverse_dict(a)
		if ad_f:
			d.update(ad_f.make_reverse_dict(a))
		for e in self.outputs:
			if e.addr and e.addr in d:
				e.mmid, f = d[e.addr]
				if f:
					e.comment = f

	def check_dup_addrs(self, io_desc):
		assert io_desc in ('inputs', 'outputs')
		addrs = [e.addr for e in getattr(self, io_desc) if e.addr]
		if len(addrs) != len(set(addrs)):
			die(2, f'{addrs}: duplicate address in transaction {io_desc}')

	# given tx size and absolute fee or fee spec, return absolute fee
	# relative fee is N+<first letter of unit name>
	def feespec2abs(self, fee_arg, tx_size):

		if type(fee_arg) is self.proto.coin_amt:
			return fee_arg

		if fee := get_obj(self.proto.coin_amt, num=fee_arg, silent=True):
			return fee

		if res := parse_fee_spec(self.proto, fee_arg):
			return self.fee_rel2abs(tx_size, float(res.amt), res.unit)

		return False

	def get_usr_fee_interactive(self, fee=None, *, desc='Starting'):
		abs_fee = None
		from ..ui import line_input
		while True:
			if fee:
				abs_fee = self.convert_and_check_fee(fee, desc)
			if abs_fee:
				if self.is_bump and not self.check_bumped_fee_ok(abs_fee):
					pass
				else:
					prompt = '{a} TX fee{b}: {c}{d} {e} ({f} {g})\n'.format(
						a = desc,
						b = (f' (after {self.cfg.fee_adjust:.2f}X adjustment)'
								if self.cfg.fee_adjust != 1 and desc.startswith('Network-estimated')
									else ''),
						c = ('', '≈')[self.fee_is_approximate],
						d = abs_fee.hl(),
						e = self.coin,
						f = pink(self.fee_abs2rel(abs_fee)),
						g = self.rel_fee_disp)
					from ..ui import keypress_confirm
					if self.cfg.yes or keypress_confirm(self.cfg, prompt+'OK?', default_yes=True):
						if self.cfg.yes:
							msg(prompt)
						return abs_fee
			fee = line_input(self.cfg, self.usr_fee_prompt)
			desc = 'User-selected'

	# we don't know fee yet, so perform preliminary check with fee == 0
	async def precheck_sufficient_funds(self, inputs_sum, sel_unspent, outputs_sum):
		if self.twuo.total < outputs_sum:
			msg(self.msg_wallet_low_coin.format(outputs_sum-inputs_sum, self.dcoin))
			return False
		if inputs_sum < outputs_sum:
			self.warn_insufficient_funds(outputs_sum - inputs_sum, self.dcoin)
			return False
		return True

	def add_output(self, coinaddr, amt, *, is_chg=False, is_vault=False, data=None):
		self.outputs.append(
			self.Output(self.proto, addr=coinaddr, amt=amt, is_chg=is_chg, is_vault=is_vault, data=data))

	def process_data_output_arg(self, arg):
		return None

	def parse_cmdline_arg(self, proto, arg_in, ad_f, ad_w):

		_pa = namedtuple('txcreate_cmdline_output', ['arg', 'mmid', 'addr', 'amt', 'data', 'is_vault'])

		if data := self.process_data_output_arg(arg_in):
			return _pa(arg_in, None, None, None, data, False)

		arg, amt = arg_in.split(',', 1) if ',' in arg_in else (arg_in, None)

		coin_addr, mmid, is_vault = (None, None, False)

		if arg == 'vault' and self.is_swap:
			is_vault = True
		elif mmid := get_obj(MMGenID, proto=proto, id_str=arg, silent=True):
			coin_addr = mmaddr2coinaddr(self.cfg, arg, ad_w, ad_f, proto)
		elif is_coin_addr(proto, arg):
			coin_addr = CoinAddr(proto, arg)
		elif is_mmgen_addrtype(proto, arg) or is_addrlist_id(proto, arg):
			if proto.base_proto_coin != 'BTC':
				die(2, f'Change addresses not supported for {proto.name} protocol')
			self.chg_autoselected = True
		else:
			die(2, f'{arg_in}: invalid command-line argument')

		return _pa(arg, mmid, coin_addr, amt, None, is_vault)

	async def get_autochg_addr(self, proto, arg, *, exclude, desc, all_addrtypes=False):
		from ..tw.addresses import TwAddresses
		al = await TwAddresses(self.cfg, proto, get_data=True)

		if all_addrtypes:
			res = al.get_change_address_by_addrtype(None, exclude=exclude, desc=desc)
			req_desc = 'of any allowed address type'
		elif obj := get_obj(MMGenAddrType, proto=proto, id_str=arg, silent=True):
			res = al.get_change_address_by_addrtype(obj, exclude=exclude, desc=desc)
			req_desc = f'of address type {arg!r}'
		else:
			res = al.get_change_address(arg, exclude=exclude, desc=desc)
			req_desc = f'from address list {arg!r}'

		if res:
			return res

		die(2, 'Tracking wallet contains no {t}addresses {d}'.format(
			t = '' if res is None else 'unused ',
			d = req_desc))

	async def process_cmdline_args(self, cmd_args, ad_f, ad_w):

		parsed_args = [self.parse_cmdline_arg(self.proto, arg, ad_f, ad_w) for arg in cmd_args]

		chg_args = [a for a in parsed_args if not (a.amt or a.data)]

		if len(chg_args) > 1:
			desc = 'requested' if self.chg_autoselected else 'listed'
			die(2, f'ERROR: More than one change address {desc} on command line')

		for a in parsed_args:
			if a.data:
				self.add_output(None, self.proto.coin_amt('0'), data=a.data)
			else:
				self.add_output(
					coinaddr = None if a.is_vault else a.addr or (
						await self.get_autochg_addr(
							self.proto,
							a.arg,
							exclude = [a.mmid for a in parsed_args if a.mmid],
							desc = 'change address')).addr,
					amt = self.proto.coin_amt(a.amt or '0'),
					is_chg = not a.amt,
					is_vault = a.is_vault)

		if self.chg_idx is None:
			die(2,
				fmt(self.msg_no_change_output.format(self.dcoin)).strip()
					if len(self.outputs) == 1 else
				'ERROR: No change output specified')

		if self.has_segwit_outputs() and not self.rpc.info('segwit_is_active'):
			die(2,
				f'{gc.proj_name} Segwit address requested on the command line, '
				'but Segwit is not active on this chain')

		if not self.nondata_outputs:
			die(2, 'At least one spending output must be specified on the command line')

		self.add_mmaddrs_to_outputs(ad_f, ad_w)

		self.check_dup_addrs('outputs')

		if self.chg_output is not None:
			if self.chg_autoselected and not self.is_swap: # swap TX, so user has already confirmed
				self.confirm_autoselected_addr(self.chg_output.mmid, 'change address')
			elif len(self.nondata_outputs) > 1:
				await self.warn_addr_used(self.proto, self.chg_output, 'change address')

	def get_addrfiles_from_cmdline(self, cmd_args):
		from ..addrfile import AddrFile
		addrfile_args = remove_dups(
			tuple(a for a in cmd_args if get_extension(a) == AddrFile.ext),
			desc = 'command line',
			edesc = 'argument',
		)
		cmd_args = tuple(a for a in cmd_args if a not in addrfile_args)
		if not self.is_swap:
			cmd_args = remove_dups(cmd_args, desc='command line', edesc='argument')
		return cmd_args, addrfile_args

	def get_addrdata_from_files(self, proto, addrfiles):
		from ..addrdata import AddrData
		from ..addrlist import AddrList
		from ..fileutil import check_infile
		ad_f = AddrData(proto)
		for addrfile in addrfiles:
			check_infile(addrfile)
			try:
				ad_f.add(AddrList(self.cfg, proto, infile=addrfile))
			except Exception as e:
				msg(f'{type(e).__name__}: {e}')
		return ad_f

	def confirm_autoselected_addr(self, mmid, desc):
		from ..ui import keypress_confirm
		keypress_confirm(
			self.cfg,
			'Using {a} as {b}. OK?'.format(
				a = mmid.hl(),
				b = 'single output address' if len(self.nondata_outputs) == 1 else desc),
			default_yes = True,
			do_exit = True)

	async def warn_addr_used(self, proto, chg, desc):
		if proto.address_reuse_ok:
			return
		from ..tw.addresses import TwAddresses
		if (await TwAddresses(self.cfg, proto, get_data=True)).is_used(chg.addr):
			from ..ui import keypress_confirm
			keypress_confirm(
				self.cfg,
				'{a} {b} {c}\n{d}'.format(
					a = yellow(f'Requested {desc}'),
					b = chg.mmid.hl() if chg.mmid else chg.addr.hl(chg.addr.view_pref),
					c = yellow('is already used!'),
					d = yellow('Address reuse harms your privacy and security. Continue anyway? (y/N): ')
				),
				complete_prompt = True,
				default_yes = False,
				do_exit = True)

	# inputs methods
	def get_unspent_nums_from_user(self, unspent):
		prompt = 'Enter a range or space-separated list of outputs to spend: '
		from ..ui import line_input
		while True:
			reply = line_input(self.cfg, prompt).strip()
			if reply:
				from ..addrlist import AddrIdxList
				selected = get_obj(AddrIdxList, fmt_str=','.join(reply.split()))
				if selected:
					if selected[-1] <= len(unspent):
						return selected
					msg(f'Unspent output number must be <= {len(unspent)}')

	def get_unspent_nums_from_inputs_opt(self, unspent):

		def do_add_msg(idx):
			uo = unspent[idx]
			mm_disp = f' ({uo.twmmid})' if uo.twmmid.type == 'mmgen' else ''
			msg('Adding input: {} {}{}'.format(idx + 1, uo.addr, mm_disp))

		def get_uo_nums():
			for addr in self.cfg.inputs.split(','):
				if is_mmgen_id(self.proto, addr):
					attr = 'twmmid'
				elif is_coin_addr(self.proto, addr):
					attr = 'addr'
				else:
					die(1, f'{addr!r}: not an MMGen ID or {self.coin} address')

				found = False
				for idx, e in enumerate(unspent):
					if getattr(e, attr) == addr:
						do_add_msg(idx)
						yield idx + 1
						found = True

				if not found:
					die(1, f'{addr!r}: address not found in tracking wallet')

		return set(get_uo_nums()) # silently discard duplicates

	def copy_inputs_from_tw(self, tw_unspent_data):
		def gen_inputs():
			for d in tw_unspent_data:
				i = self.Input(
					self.proto,
					**{attr:getattr(d, attr) for attr in d.__dict__ if attr in self.Input.tw_copy_attrs})
				if d.twmmid.type == 'mmgen':
					i.mmid = d.twmmid # twmmid -> mmid
				yield i
		self.inputs = type(self.inputs)(self, list(gen_inputs()))

	async def get_funds_available(self, fee, outputs_sum):
		in_sum = self.sum_inputs()
		out_sum = outputs_sum + fee
		return self._funds_available(
			in_sum >= out_sum,
			# CoinAmt must be non-negative, so cannot use abs():
			in_sum - out_sum if in_sum >= out_sum else out_sum - in_sum)

	async def get_inputs(self, outputs_sum):
		sel_nums = (
			self.get_unspent_nums_from_inputs_opt if self.cfg.inputs else
			self.get_unspent_nums_from_user
		)(self.twuo.data)

		msg('Selected {}{}: {}'.format(
			self.twuo.item_desc,
			suf(sel_nums),
			' '.join(str(n) for n in sel_nums)))
		sel_unspent = MMGenList(self.twuo.data[i-1] for i in sel_nums)

		if not await self.precheck_sufficient_funds(
				sum(s.amt for s in sel_unspent),
				sel_unspent,
				outputs_sum):
			return False

		self.copy_inputs_from_tw(sel_unspent)  # makes self.inputs
		return True

	async def network_fee_disp(self):
		res = await self.get_rel_fee_from_network()
		return pink(
			'N/A' if res.fee is None else
			self.network_fee_to_unit_disp(res))

	async def get_fee(self, fee, outputs_sum, start_fee_desc):

		if fee:
			self.usr_fee = self.get_usr_fee_interactive(fee, desc=start_fee_desc)
		else:
			res = await self.get_rel_fee_from_network()
			self.usr_fee = self.get_usr_fee_interactive(
				None if res.fee is None else self.fee_est2abs(res),
				desc = self.network_estimated_fee_label)

		funds = await self.get_funds_available(self.usr_fee, outputs_sum)

		if funds.is_positive:
			p = self.final_inputs_ok_msg(funds.amt)
			from ..ui import keypress_confirm
			if self.cfg.yes or keypress_confirm(self.cfg, p+'. OK?', default_yes=True):
				if self.cfg.yes:
					msg(p)
				return funds.amt
		else:
			self.warn_insufficient_funds(funds.amt, self.coin)

	def _non_wallet_addr_confirm(self, message):
		from ..ui import confirm_or_raise
		confirm_or_raise(
			cfg = self.cfg,
			message = yellow(message),
			action = 'Are you sure this is what you want?')

	async def create(self, cmd_args, *, locktime=None, do_info=False, caller='txcreate'):

		assert isinstance(locktime, (int, type(None))), 'locktime must be of type int'

		from ..tw.unspent import TwUnspentOutputs

		if self.cfg.comment_file:
			self.add_comment(infile=self.cfg.comment_file)

		if not do_info:
			cmd_args, addrfile_args = self.get_addrfiles_from_cmdline(cmd_args)
			if self.is_swap:
				cmd_args = await self.process_swap_cmdline_args(cmd_args, addrfile_args)
			from ..rpc import rpc_init
			self.rpc = await rpc_init(self.cfg, self.proto)
			from ..addrdata import TwAddrData
			await self.process_cmdline_args(
				cmd_args,
				self.get_addrdata_from_files(self.proto, addrfile_args),
				await TwAddrData(self.cfg, self.proto, twctl=self.twctl))

		if not self.is_bump:
			self.twuo = await TwUnspentOutputs(
				self.cfg,
				self.proto,
				minconf = self.cfg.minconf,
				addrs = await self.get_input_addrs_from_inputs_opt())
			await self.twuo.get_data()
			self.twctl = self.twuo.twctl

		from ..ui import do_license_msg
		do_license_msg(self.cfg)

		if not (self.is_bump or self.cfg.inputs):
			await self.twuo.view_filter_and_sort()

		if not self.is_bump:
			self.twuo.display_total()

		if do_info:
			del self.twctl
			del self.twuo.twctl
			import sys
			sys.exit(0)

		outputs_sum = self.sum_outputs()

		msg('Total amount to spend: {}'.format(
			f'{outputs_sum.hl()} {self.dcoin}' if outputs_sum else 'Unknown'
		))

		while True:
			if not await self.get_inputs(outputs_sum):
				continue
			if self.is_swap:
				fee_hint = await self.update_vault_output(
					self.vault_output.amt or self.sum_inputs(),
					deduct_est_fee = self.vault_output == self.chg_output)
			else:
				await self.set_gas()
				fee_hint = None
			desc = 'User-selected' if self.cfg.fee else 'Recommended' if fee_hint else None
			if (funds_left := await self.get_fee(
					self.cfg.fee or fee_hint,
					outputs_sum,
					desc)) is not None:
				break

		self.check_non_mmgen_inputs(caller=caller)

		self.update_change_output(funds_left)

		self.check_chg_addr_is_wallet_addr()

		if not self.cfg.yes:
			self.add_comment()  # edits an existing comment

		if self.is_swap:
			import time
			if time.time() > self.swap_quote_refresh_time + self.swap_quote_refresh_timeout:
				await self.update_vault_output(self.vault_output.amt)

		await self.create_serialized(locktime=locktime) # creates self.txid too

		self.add_timestamp()
		self.add_blockcount()
		self.chain = self.proto.chain_name
		self.check_fee()

		self.cfg._util.qmsg('Transaction successfully created')

		if self.is_bump:
			return

		from . import UnsignedTX
		new = UnsignedTX(cfg=self.cfg, data=self.__dict__, automount=self.cfg.autosign)

		if not self.cfg.yes:
			new.info.view_with_prompt('View transaction details?')

		del new.twuo.twctl
		return new



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.status: THORChain transaction status class
"""

from ....util import msg, gmsg
from ....tx import status as TxBase

class Status(TxBase.Status):

	async def display(self, *, idx=''):

		try:
			self.tx.rpc.get_tx_info(self.tx.coin_txid)
		except Exception as e:
			msg(f'{type(e).__name__}: {e}')
			return 2

		gmsg('Transaction is in blockchain')
		return 0



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.include.cfg: configuration data for cmdtest.py
"""

from ...include.common import cfg
from .common import pwfile, hincog_fn, incog_id_fn, randbool

cmd_groups_altcoin = ['ref_altcoin', 'autosign', 'ethdev', 'xmrwallet', 'xmr_autosign']

cmd_groups_dfl = {
	'misc':               ('CmdTestMisc',              {}),
	'opts':               ('CmdTestOpts',              {'full_data': True}),
	'cfgfile':            ('CmdTestCfgFile',           {'full_data': True}),
	'help':               ('CmdTestHelp',              {'full_data': True}),
	'main':               ('CmdTestMain',              {'full_data': True}),
	'conv':               ('CmdTestWalletConv',        {'is3seed': True, 'modname': 'wallet'}),
	'ref':                ('CmdTestRef',               {}),
	'ref3':               ('CmdTestRef3Seed',          {'is3seed': True, 'modname': 'ref_3seed'}),
	'ref3_addr':          ('CmdTestRef3Addr',          {'is3seed': True, 'modname': 'ref_3seed'}),
	'ref3_pw':            ('CmdTestRef3Passwd',        {'is3seed': True, 'modname': 'ref_3seed'}),
	'ref_altcoin':        ('CmdTestRefAltcoin',        {}),
	'seedsplit':          ('CmdTestSeedSplit',         {}),
	'tool':               ('CmdTestTool',              {'full_data': True}),
	'input':              ('CmdTestInput',             {}),
	'output':             ('CmdTestOutput',            {'modname': 'misc', 'full_data': True}),
	'autosign_clean':     ('CmdTestAutosignClean',     {'modname': 'autosign'}),
	'autosign':           ('CmdTestAutosign',          {}),
	'autosign_automount': ('CmdTestAutosignAutomount', {'modname': 'automount'}),
	'autosign_eth':       ('CmdTestAutosignETH',       {'modname': 'automount_eth'}),
	'regtest':            ('CmdTestRegtest',           {}),
	'swap':               ('CmdTestSwap',              {}),
	'ethswap':            ('CmdTestEthSwap',           {}),
	# 'chainsplit':         ('CmdTestChainsplit',      {}),
	'ethdev':             ('CmdTestEthdev',            {}),
	'ethbump':            ('CmdTestEthBump',           {}),
	'rune':               ('CmdTestRune',              {}),
	'runeswap':           ('CmdTestRuneSwap',          {}),
	'xmrwallet':          ('CmdTestXMRWallet',         {}),
	'xmr_autosign':       ('CmdTestXMRAutosign',       {}),
}

cmd_groups_extra = {
	'ethswap_eth':            ('CmdTestEthSwapEth',           {'modname': 'ethswap'}),
	'ethbump_ltc':            ('CmdTestEthBumpLTC',           {'modname': 'ethbump'}),
	'runeswap_rune':          ('CmdTestRuneSwapRune',         {'modname': 'runeswap'}),
	'dev':                    ('CmdTestDev',                  {'modname': 'misc'}),
	'regtest_legacy':         ('CmdTestRegtestBDBWallet',     {'modname': 'regtest'}),
	'autosign_btc':           ('CmdTestAutosignBTC',          {'modname': 'autosign'}),
	'autosign_live':          ('CmdTestAutosignLive',         {'modname': 'autosign'}),
	'autosign_live_simulate': ('CmdTestAutosignLiveSimulate', {'modname': 'autosign'}),
	'create_ref_tx':          ('CmdTestRefTX',                {'modname': 'misc', 'full_data': True}),
}

cfgs = { # addr_idx_lists (except 31, 32, 33, 34) must contain exactly 8 addresses
	'1': {
		'wpasswd':       'Dorian-α',
		'kapasswd':      'Grok the blockchain',
		'addr_idx_list': '12,99,5-10,5,12',
		'dep_generators': {
			pwfile:        'walletgen',
			'mmdat':       'walletgen',
			'addrs':       'addrgen',
			'rawtx':       'txcreate',
			'txbump':      'txbump',
			'sigtx':       'txsign',
			'mmwords':     'export_mnemonic',
			'mmseed':      'export_seed',
			'mmhex':       'export_hex',
			'mmincog':     'export_incog',
			'mmincox':     'export_incog_hex',
			hincog_fn:     'export_incog_hidden',
			incog_id_fn:   'export_incog_hidden',
			'akeys.mmenc': 'keyaddrgen'
		},
	},
	'2': {
		'wpasswd':       'Hodling away',
		'addr_idx_list': '37,45,3-6,22-23',
		'seed_len':      128,
		'dep_generators': {
			'mmdat':   'walletgen2',
			'addrs':   'addrgen2',
			'rawtx':   'txcreate2',
			'sigtx':   'txsign2',
			'mmwords': 'export_mnemonic2',
		},
	},
	'3': {
		'wpasswd':       'Major miner',
		'addr_idx_list': '73,54,1022-1023,2-5',
		'dep_generators': {
			'mmdat': 'walletgen3',
			'addrs': 'addrgen3',
			'rawtx': 'txcreate3',
			'sigtx': 'txsign3'
		},
	},
	'4': {
		'wpasswd':       'Hashrate good',
		'addr_idx_list': '63,1004,542-544,7-9',
		'seed_len':      192,
		'dep_generators': {
			'mmdat':   'walletgen4',
			'mmbrain': 'walletgen4',
			'addrs':   'addrgen4',
			'rawtx':   'txcreate4',
			'sigtx':   'txsign4',
			'txdo':    'txdo4',
		},
		'bw_filename': 'brainwallet.mmbrain',
		'bw_params':   '192,1',
	},
	'5': {
		'wpasswd':     'My changed password-α',
		'hash_preset': '2',
		'dep_generators': {
			'mmdat': 'passchg',
			pwfile:  'passchg',
		},
	},
	'6': {
		'seed_len':       128,
		'seed_id':        'FE3C6545',
		'ref_bw_seed_id': '33F10310',
		'wpasswd':        'reference password',
		'kapasswd':      '',
		'dep_generators': {
			'mmdat':       'ref_walletgen_brain_1',
			pwfile:        'ref_walletgen_brain_1',
			'addrs':       'refaddrgen_1',
			'akeys.mmenc': 'refkeyaddrgen_1'
		},
	},
	'7': {
		'seed_len':       192,
		'seed_id':        '1378FC64',
		'ref_bw_seed_id': 'CE918388',
		'wpasswd':        'reference password',
		'kapasswd':      '',
		'dep_generators': {
			'mmdat':       'ref_walletgen_brain_2',
			pwfile:        'ref_walletgen_brain_2',
			'addrs':       'refaddrgen_2',
			'akeys.mmenc': 'refkeyaddrgen_2'
		},
	},
	'8': {
		'seed_len':       256,
		'seed_id':        '98831F3A',
		'ref_bw_seed_id': 'B48CD7FC',
		'wpasswd':        'reference password',
		'kapasswd':      '',
		'dep_generators': {
			'mmdat':       'ref_walletgen_brain_3',
			pwfile:        'ref_walletgen_brain_3',
			'addrs':       'refaddrgen_3',
			'akeys.mmenc': 'refkeyaddrgen_3'
		},
	},
	'9': {
		'tool_enc_infn': 'tool_encrypt.in',
		'dep_generators': {
			'tool_encrypt.in':       'tool_encrypt',
			'tool_encrypt.in.mmenc': 'tool_encrypt',
		},
	},
	'11': {}, # wallet
	'12': {}, # wallet
	'13': {}, # wallet
	'14': {
		'kapasswd':      'Maxwell',
		'wpasswd':       'The Halving',
		'addr_idx_list': '61,998,502-504,7-9',
		'seed_len':      256,
		'dep_generators': {
			'mmdat':       'walletgen14',
			'addrs':       'addrgen14',
			'akeys.mmenc': 'keyaddrgen14',
		},
	},
	'15': {
		'wpasswd':       'Dorian-α',
		'kapasswd':      'Grok the blockchain',
		'addr_idx_list': '12,99,5-10,5,12',
		'dep_generators': {
			pwfile:       'walletgen_dfl_wallet',
			'addrs':      'addrgen_dfl_wallet',
			'rawtx':      'txcreate_dfl_wallet',
			'sigtx':      'txsign_dfl_wallet',
			'mmseed':     'export_seed_dfl_wallet',
			'del_dw_run': 'delete_dfl_wallet',
		},
	},
	'16': {
		'wpasswd':     'My changed password',
		'hash_preset': '2',
		'dep_generators': {
			pwfile: 'passchg_dfl_wallet',
		},
	},
	'17': {}, # regtest
	'18': {}, # autosign
	'19': {'wpasswd': 'abc'},
	'20': {
		'wpasswd':       'Vsize it',
		'addr_idx_list': '1-8',
		'seed_len':      256,
		'dep_generators': {
			'mmdat': 'walletgen5',
			'addrs': 'addrgen5',
			'rawtx': 'txcreate5',
			'sigtx': 'txsign5',
		},
	},
	'21': {
		'wpasswd':       'Vsize it',
		'addr_idx_list': '1-8',
		'seed_len':      256,
		'dep_generators': {
			'mmdat': 'walletgen6',
			'addrs': 'addrgen6',
			'rawtx': 'txcreate6',
			'sigtx': 'txsign6',
		},
	},
	'22': {}, # ethdev
	'23': {}, # seedsplit
	'26': {}, # ref_3seed
	'27': {}, # ref_3seed
	'28': {}, # ref_3seed
	'29': {}, # xmrwallet
	'31': {}, # ref_tx
	'32': {}, # ref_tx
	'33': {}, # ref_tx
	'34': {}, # ref_tx
	'37': {}, # swap
	'38': {}, # autosign_clean
	'39': {}, # xmr_autosign
	'40': {}, # cfgfile
	'41': {}, # opts
	'42': {}, # ethbump
	'43': {}, # ethbump_ltc
	'47': {}, # ethswap
	'48': {}, # ethswap_eth
	'49': {}, # autosign_automount
	'50': {}, # rune
	'57': {}, # runeswap
	'58': {}, # runeswap_rune
	'59': {}, # autosign_eth
	'99': {}, # dummy
}

def fixup_cfgs():
	import os

	for k in cfgs:
		cfgs[k]['tmpdir'] = os.path.join('test', 'tmp', str(k))

	for src, target in (
			('6', '11'),
			('7', '12'),
			('8', '13'),
			('6', '26'),
			('7', '27'),
			('8', '28')):
		cfgs[target].update(cfgs[src])
		cfgs[target]['tmpdir'] = os.path.join('test', 'tmp', target)

	for k in cfgs:
		cfgs[k]['segwit'] = randbool() if cfg.segwit_random else bool(cfg.segwit or cfg.bech32)

	if cfg.debug_utf8:
		for k in cfgs:
			cfgs[k]['tmpdir'] += '-α'

fixup_cfgs()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txbump: Create, and optionally send and sign, a replacement transaction
              on supporting networks
"""

from .cfg import gc, Config
from .util import msg, msg_r, die, async_run
from .color import green

opts_data = {
	'filter_codes': ['-'],
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': """
                Create, and optionally send and sign, a replacement transaction
                on supporting networks
		 """,
		'usage2':   (
			f'[opts] [{gc.proj_name} TX file] [seed source] ...',
			f'[opts] {{u_args}} [{gc.proj_name} TX file] [seed source] ...',
		),
		'options': """
			-- -h, --help             Print this help message
			-- --, --longhelp         Print help message for long (global) options
			-- -a, --autosign         Bump the most recent transaction created and sent with
			+                         the --autosign option. The removable device is mounted
			+                         and unmounted automatically.  The transaction file
			+                         argument must be omitted.  Note that only sent trans-
			+                         actions may be bumped with this option.  To redo an
			+                         unsent --autosign transaction, first delete it using
			+                         ‘mmgen-txsend --abort’ and then create a new one
			-- -b, --brain-params=l,p Use seed length 'l' and hash preset 'p' for
			+                         brainwallet input
			-- -c, --comment-file=  f Source the transaction's comment from file 'f'
			-- -d, --outdir=        d Specify an alternate directory 'd' for output
			-- -e, --echo-passphrase  Print passphrase to screen when typing it
			-- -f, --fee=           f Transaction fee, as a decimal {cu} amount or as
			+                         {fu} (an integer followed by {fl}).
			+                         See FEE SPECIFICATION below.
			e- -g, --gas=N            Set the gas limit (see GAS LIMIT below)
			-- -H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
			+                        'f' at offset 'o' (comma-separated)
			-- -i, --in-fmt=        f Input is from wallet format 'f' (see FMT CODES below)
			-- -l, --seed-len=      l Specify wallet seed length of 'l' bits. This option
			+                         is required only for brainwallet and incognito inputs
			+                         with non-standard (< {dsl}-bit) seed lengths.
			-- -k, --keys-from-file=f Provide additional keys for non-{pnm} addresses
			-- -K, --keygen-backend=n Use backend 'n' for public key generation.  Options
			+                         for {coin_id}: {kgs}
			-- -M, --mmgen-keys-from-file=f Provide keys for {pnm} addresses in a key-
			+                         address file (output of '{pnl}-keygen'). Permits
			+                         online signing without an {pnm} seed source. The
			+                         key-address file is also used to verify {pnm}-to-{cu}
			+                         mappings, so the user should record its checksum.
			b- -o, --output-to-reduce=o Deduct the fee from output 'o' (an integer, or 'c'
			+                         for the transaction's change output, if present)
			-- -O, --old-incog-fmt    Specify old-format incognito input
			-- -p, --hash-preset=   p Use the scrypt hash parameters defined by preset 'p'
			+                         for password hashing (default: '{gc.dfl_hash_preset}')
			-- -P, --passwd-file=   f Get {pnm} wallet passphrase from file 'f'
			-- -q, --quiet            Suppress warnings; overwrite files without prompting
			-- -s, --send             Sign and send the transaction (the default if seed
			+                         data is provided)
			-- -T, --txhex-idx=N      Send only part ‘N’ of a multi-part transaction.
			+                         Indexing begins with one.
			-- -v, --verbose          Produce more verbose output
			e- -w, --wait             Wait for transaction confirmation
			-- -W, --allow-non-wallet-swap Allow signing of swap transactions that send funds
			+                         to non-wallet addresses
			-- -x, --proxy=P          Fetch the swap quote via SOCKS5h proxy ‘P’ (host:port).
			+                         Use special value ‘env’ to honor *_PROXY environment
			+                         vars instead.
			-- -y, --yes              Answer 'yes' to prompts, suppress non-essential output
			-- -z, --show-hash-presets Show information on available hash presets
""",
	'notes': """

With --autosign, the TX file argument is omitted, and the last submitted TX
file on the removable device will be used.

If no outputs are specified, the original outputs will be used for the
replacement transaction, otherwise a new transaction will be created with the
outputs listed on the command line.  The syntax for the output arguments is
identical to that of ‘mmgen-txcreate’.

The user should take care to select a fee sufficient to ensure the original
transaction is replaced in the mempool.

When bumping a swap transaction, the swap protocol’s quote server on the
Internet must be reachable either directly or via the SOCKS5 proxy specified
with the --proxy option.  To improve privacy, it’s recommended to proxy
requests to the quote server via Tor or some other anonymity network.

{g}{F}{s}
Seed source files must have the canonical extensions listed in the 'FileExt'
column below:

{f}
"""
	},
	'code': {
		'usage': lambda cfg, proto, help_notes, s: s.format(
			u_args = help_notes('txcreate_args')),
		'options': lambda cfg, help_notes, proto, s: s.format(
			cfg     = cfg,
			gc      = gc,
			pnm     = gc.proj_name,
			pnl     = gc.proj_name.lower(),
			fu      = help_notes('rel_fee_desc'),
			fl      = help_notes('fee_spec_letters', use_quotes=True),
			kgs     = help_notes('keygen_backends'),
			coin_id = help_notes('coin_id'),
			dsl     = help_notes('dfl_seed_len'),
			cu      = proto.coin),
		'notes': lambda help_mod, help_notes, s: s.format(
			g       = help_notes('gas_limit', target=None),
			F       = help_notes('fee'),
			s       = help_mod('txsign'),
			f       = help_notes('fmt_codes')),
	}
}

cfg = Config(opts_data=opts_data)

from .tx import CompletedTX, BumpTX, UnsignedTX, OnlineSignedTX
from .tx.keys import TxKeys, pop_seedfiles, get_keylist, get_keyaddrlist

seedfiles = pop_seedfiles(cfg, ignore_dfl_wallet=not cfg.send, empty_ok=not cfg.send)

if cfg.autosign:
	if cfg.send:
		die(1, '--send cannot be used together with --autosign')
else:
	tx_file = cfg._args.pop()
	from .fileutil import check_infile
	check_infile(tx_file)

from .ui import do_license_msg
do_license_msg(cfg)

silent = cfg.yes and cfg.fee is not None and cfg.output_to_reduce is not None

async def main():

	if cfg.autosign:
		from .tx.util import mount_removable_device
		from .autosign import Signable
		asi = mount_removable_device(cfg)
		si = Signable.automount_transaction(asi)
		if si.unsigned or si.unsent:
			state = 'unsigned' if si.unsigned else 'unsent'
			die(1,
				'Only sent transactions can be bumped with --autosign.  Instead of bumping\n'
				f'your {state} transaction, abort it with ‘mmgen-txsend --abort’ and create\n'
				'a new one.')
		orig_tx = await si.get_last_created()
		sign_and_send = False
	else:
		orig_tx = await CompletedTX(cfg=cfg, filename=tx_file)
		kl = get_keylist(cfg)
		kal = get_keyaddrlist(cfg, orig_tx.proto)
		sign_and_send = any([seedfiles, kl, kal])

	if not silent:
		msg(green('ORIGINAL TRANSACTION'))
		msg(orig_tx.info.format(terse=True))

	from .tw.ctl import TwCtl
	tx = BumpTX(
		cfg  = cfg,
		data = orig_tx.__dict__,
		automount = cfg.autosign,
		check_sent = cfg.autosign or sign_and_send,
		new_outputs = bool(cfg._args),
		twctl = await TwCtl(cfg, orig_tx.proto) if orig_tx.proto.tokensym else None)

	if tx.new_outputs:
		await tx.create(cfg._args, caller='txdo' if sign_and_send else 'txcreate')
	else:
		await tx.create_feebump(silent=silent)

	if not silent:
		msg(green('\nREPLACEMENT TRANSACTION:'))
		msg_r(tx.info.format(terse=True))

	if sign_and_send:
		tx2 = UnsignedTX(cfg=cfg, data=tx.__dict__)
		if tx3 := await tx2.sign(
				TxKeys(cfg, tx2, seedfiles=seedfiles, keylist=kl, keyaddrlist=kal).keys):
			tx4 = await OnlineSignedTX(cfg=cfg, data=tx3.__dict__)
			tx4.file.write(ask_write=False)
			await tx4.send(cfg, asi if cfg.autosign else None)
		else:
			die(2, 'Transaction could not be signed')
	else:
		tx.file.write(
			outdir                = asi.txauto_dir if cfg.autosign else None,
			ask_write             = not cfg.yes,
			ask_write_default_yes = False,
			ask_overwrite         = not cfg.yes)

async_run(main())



#!/usr/bin/env python3

"""
test.modtest_d.xmrseed: Monero mnemonic unit test for the MMGen suite
"""

from mmgen.util import ymsg
from mmgen.xmrseed import xmrseed
from ..include.common import cfg, vmsg

class unit_tests:

	vectors = ( # private keys are reduced
		(
			'148d78d2aba7dbca5cd8f6abcfb0b3c009ffbdbea1ff373d50ed94d78286640e', # Monero repo
			'velvet lymph giddy number token physics poetry unquoted nibs useful sabotage limits benches ' +
			'lifestyle eden nitrogen anvil fewest avoid batch vials washing fences goat unquoted',
		), (
			'e8164dda6d42bd1e261a3406b2038dcbddadbeefdeadbeefdeadbeefdeadbe0f',
			'viewpoint donuts ardent template unveil agile meant unafraid urgent athlete rustled mime azure ' +
			'jaded hawk baby jagged haystack baby jagged haystack ramped oncoming point template'
		), (
			'6900dea9753f5c7ced87b53bdcfb109a8417bca6a2797a708194157b227fb60b',
			'criminal bamboo scamper gnaw limits womanly wrong tuition birth mundane donuts square cohesive ' +
			'dolphin titans narrate fuel saved wrap aloof magically mirror together update wrap'
		), (
			'0000000000000000000000000000000000000000000000000000000000000001',
			'abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey ' +
			'abbey abbey abbey abbey abbey bamboo jaws jerseys abbey'
		), (
			'1c95988d7431ecd670cf7d73f45befc6feffffffffffffffffffffffffffff0f',
			'powder directed sayings enmity bacon vapidly entrance bumper noodles iguana sleepless nasty flying ' +
			'soil software foamy solved soggy foamy solved soggy jury yawning ankle solved'
		), (
			'2c94988d7431ecd670cf7d73f45befc6feffffffffffffffffffffffffffff0f',
			'memoir apart olive enmity bacon vapidly entrance bumper noodles iguana sleepless nasty flying soil ' +
			'software foamy solved soggy foamy solved soggy jury yawning ankle foamy'
		), (
			'4bb0288c9673b69fa68c2174851884abbaaedce6af48a03bbfd25e8cd0364102',
			'rated bicycle pheasants dejected pouch fizzle shipped rash citadel queen avatar sample muzzle mews ' +
			'jagged origin yeti dunes obtains godfather unbending pastry vortex washing citadel'
		), (
			'4bb0288c9673b69fa68c2174851884abbaaedce6af48a03bbfd25e8cd0364100',
			'rated bicycle pheasants dejected pouch fizzle shipped rash citadel queen avatar sample muzzle mews ' +
			'jagged origin yeti dunes obtains godfather unbending kangaroo auctions audio citadel'
		), (
			'1d95988d7431ecd670cf7d73f45befc6feffffffffffffffffffffffffffff0e',
			'pram distance scamper enmity bacon vapidly entrance bumper noodles iguana sleepless nasty flying ' +
			'soil software foamy solved soggy foamy solved soggy hashing mullet onboard solved'
		), (
			'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff0f',
			'foamy solved soggy foamy solved soggy foamy solved soggy foamy solved soggy foamy solved soggy ' +
			'foamy solved soggy foamy solved soggy jury yawning ankle soggy'
		),
	)

	@property
	def _use_monero_python(self):
		if not hasattr(self, '_use_monero_python_'):
			try:
				from monero.wordlists.english import English
				self.wl = English()
			except ImportError:
				self._use_monero_python_ = False
				ymsg('Warning: unable to import monero-python, skipping external library checks')
			else:
				self._use_monero_python_ = True
		return self._use_monero_python_

	def wordlist(self, name, ut, desc='Monero wordlist'):
		xmrseed().check_wordlist(cfg)
		return True

	def fromhex(self, name, ut, desc='fromhex() method'):
		b = xmrseed()
		vmsg('Checking seed to mnemonic conversion:')
		for privhex, chk in self.vectors:
			vmsg(f'    {chk}')
			chk = tuple(chk.split())
			res = b.fromhex(privhex)
			if self._use_monero_python:
				mp_chk = tuple(self.wl.encode(privhex).split())
				assert res == mp_chk, f'check failed:\nres: {res}\nchk: {chk}'
			assert res == chk, f'check failed:\nres: {res}\nchk: {chk}'
		return True

	def tohex(self, name, ut, desc='tohex() method'):
		b = xmrseed()
		vmsg('Checking mnemonic to seed conversion:')
		for chk, words in self.vectors:
			vmsg(f'    {chk}')
			res = b.tohex(words.split())
			if self._use_monero_python:
				mp_chk = self.wl.decode(words)
				assert res == mp_chk, f'check failed:\nres: {res}\nchk: {mp_chk}'
			assert res == chk, f'check failed:\nres: {res}\nchk: {chk}'
		return True

	def errors(self, name, ut, desc='error handling'):

		bad_chksum_mn = ('abbey ' * 21 + 'bamboo jaws jerseys donuts').split()
		bad_word_mn = "admire zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo".split()
		bad_seed = 'deadbeef'
		good_mn = self.vectors[0][1].split()
		good_hex = self.vectors[0][0]
		bad_len_mn = good_mn[:22]

		b = xmrseed()
		th = b.tohex
		fh = b.fromhex

		hse = 'HexadecimalStringError'
		sle = 'SeedLengthError'
		ase = 'AssertionError'
		mne = 'MnemonicError'

		ut.process_bad_data((
			('hex',               hse, 'not a hexadecimal',     lambda: fh('xx')),
			('seed len',          sle, 'invalid seed byte len', lambda: fh(bad_seed)),
			('mnemonic type',     ase, 'must be list',          lambda: th('string')),
			('pad arg (fromhex)', ase, "invalid 'pad' arg",     lambda: fh(good_hex, pad=23)),
			('pad arg (tohex)',   ase, "invalid 'pad' arg",     lambda: th(good_mn, pad=23)),
			('word',              mne, "not in Monero",         lambda: th(bad_word_mn)),
			('checksum',          mne, "checksum",              lambda: th(bad_chksum_mn)),
			('seed phrase len',   mne, "phrase len",            lambda: th(bad_len_mn)),
		))

		return True



#!/usr/bin/env python3

"""
test.modtest_d.testdep: test dependency unit tests for the MMGen suite
"""

import os
from subprocess import run, DEVNULL

from mmgen.util import ymsg, bmsg
from ..include.common import cfg, get_ethkey

sec = 'deadbeef' * 8

class unit_tests:

	altcoin_deps = ('pycoin', 'monero_python', 'keyconv', 'zcash_mini', 'eth_keys', 'ssh_socks_proxy')
	win_skip = ('losetup', 'zcash_mini', 'sudo')
	mac_skip = ('losetup',)

	def sudo(self, name, ut):
		from mmgen.util import have_sudo
		if have_sudo():
			return True
		else:
			ymsg(f'To run the test suite, please enable sudo without password for user ‘{os.getenv("USER")}’')
			return False

	def losetup(self, name, ut):
		os.stat('/dev/loop0')
		for cmd in ('/sbin/losetup', '/usr/sbin/losetup', 'losetup'):
			try:
				run([cmd, '-f'], check=True, stdout=DEVNULL)
				break
			except:
				if cmd == 'losetup':
					raise
		return True

	def pycoin(self, name, ut):
		from pycoin.networks.registry import network_for_netcode as nfnc
		network = nfnc('btc')
		key = network.keys.private(secret_exponent=int(sec, 16), is_compressed=True)
		hash160_c = key.hash160(is_compressed=True)
		network.address.for_p2pkh_wit(hash160_c)
		return True

	def monero_python(self, name, ut):
		from mmgen.util2 import load_cryptodome
		load_cryptodome()
		from monero.seed import Seed
		Seed('deadbeef' * 8).public_address()
		return True

	def keyconv(self, name, ut):
		run(['keyconv', '-G', 'ltc'], stdout=DEVNULL, stderr=DEVNULL, check=True)
		return True

	def zcash_mini(self, name, ut):
		run(['zcash-mini'], stdout=DEVNULL, check=True)
		return True

	def eth_keys(self, name, ut):
		try:
			from eth_keys import keys
			return True
		except ImportError:
			if get_ethkey():
				return True
		ymsg('Neither the ‘eth-keys’ package nor the Parity ‘ethkey’ executable '
			'could be found on the system!')

	def ssh_socks_proxy(self, name, ut):
		from test.cmdtest_d.include.proxy import TestProxy
		return TestProxy(None, cfg)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.unsigned: THORChain unsigned transaction class
"""

from ....tx import unsigned as TxBase
from ....obj import CoinTxID, NonNegativeInt
from ....addr import CoinAddr

from ...vm.tx.unsigned import Unsigned as VmUnsigned

from .completed import Completed

class Unsigned(VmUnsigned, Completed, TxBase.Unsigned):

	def parse_txfile_serialized_data(self):
		d = self.serialized
		self.txobj = {
			'from':           CoinAddr(self.proto, d['from']),
			'to':             CoinAddr(self.proto, d['to']) if d['to'] else None,
			'amt':            self.proto.coin_amt(d['amt']),
			'gas':            NonNegativeInt(d['gas']),
			'account_number': NonNegativeInt(d['account_number']),
			'sequence':       NonNegativeInt(d['sequence']),
			'chain_id':       d['chain_id']}

	async def do_sign(self, o, wif):
		if self.is_swap:
			from .protobuf import swap_tx_parms, build_swap_tx as build_tx
			parms = swap_tx_parms(
				o['from'],
				o['amt'],
				o['gas'],
				o['account_number'],
				o['sequence'],
				self.swap_memo,
				wifkey = wif)
		else:
			from .protobuf import send_tx_parms, build_tx
			parms = send_tx_parms(
				o['from'],
				o['to'],
				o['amt'],
				o['gas'],
				o['account_number'],
				o['sequence'],
				wifkey = wif)

		tx = build_tx(self.cfg, self.proto, parms)
		self.serialized = bytes(tx).hex()
		self.coin_txid = CoinTxID(tx.txid)
		tx.verify_sig(self.proto, o['account_number'])

class AutomountUnsigned(TxBase.AutomountUnsigned, Unsigned):
	pass
